"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from rse.main import Encyclopedia
from rse.exceptions import RepoNotFoundError


def main(args, extra):

    # Create a research software encyclopedia
    client = Encyclopedia(config_file=args.config_file, database=args.database)

    # If a file is defined, we import from it
    if args.file is not None:
        try:
            if args.type[0] == "criteria":
                client.import_criteria_annotation(
                    input_file=args.file, username=args.username
                )
            elif args.type[0] == "taxonomy":
                client.import_taxonomy_annotation(
                    input_file=args.file, username=args.username
                )
        except Exception as exc:
            print(exc)

    # The type is either criteria or taxonomy
    else:
        try:
            client.annotate(
                username=args.username,
                atype=args.type[0],
                unseen_only=not args.all_repos,
                repo=args.repo,
                save=True,
            )
        except RepoNotFoundError as exc:
            print(exc)



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from rse.main import Encyclopedia
from rse.logger import bot


def main(args, extra):

    # Create a queue object, run the command to match to an executor
    enc = Encyclopedia(config_file=args.config_file, database=args.database)

    # Case 1: empty list indicates listing all
    if not args.parser:
        bot.table(enc.list())
    else:
        # Each in the list can be a full executor or a uid
        for parser in args.parser:
            bot.table(enc.list(parser))



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from rse.exceptions import (
    DirectoryNotFoundError,
    MultipleReposExistError,
    RepoNotFoundError,
    NoReposError,
)
from rse.utils.file import (
    write_json,
    write_file,
    mkdir_p,
    read_json,
    read_file,
    recursive_find,
    get_latest_modified,
)
from rse.main.database.base import Database
from rse.main.parsers import get_parser
from rse.main.parsers.base import ParserBase
from glob import glob
import logging
import shutil
import os
import re

bot = logging.getLogger("rse.main.database.filesystem")


class FileSystemDatabase(Database):
    """A FileSystemDatabase writes raw json to files to a database.
    """

    database = "filesystem"

    def __init__(self, config_dir, config=None, **kwargs):
        """init for the filesystem ensures that the base folder (named 
           according to the studyid) exists.
        """
        self.config = config
        self.create_database(config_dir)

    def create_database(self, config_dir):
        """Create the database. The parent folder must exist.
        """
        self.data_base = os.path.abspath(os.path.join(config_dir, "database"))
        if not os.path.exists(config_dir):
            raise DirectoryNotFoundError(
                config_dir, "must exist to create database there"
            )
        if not os.path.exists(self.data_base):
            mkdir_p(self.data_base)

    # Global

    def clear(self):
        """clear (delete) all software repositories.
        """
        for parser_dir in self.iter_parsers(fullpath=True):
            if os.path.exists(parser_dir):
                bot.info(f"Removing {parser_dir}")
                shutil.rmtree(parser_dir)
        return True

    # Get, delete, etc. only require uid

    def exists(self, uid):
        """Determine if a repo exists.
        """
        try:
            self.get(uid, exact=True)
            return True
        except:
            return False

    def add(self, uid):
        """Add a new software repository to the database.
        """
        if uid:
            parser = get_parser(uid, config=self.config)
            data = parser.get_metadata()

            # If it's a parser handoff
            if isinstance(data, ParserBase):
                parser = data
                data = parser.data

            if data:
                bot.info(f"{parser.uid} was added to the the database.")
                return SoftwareRepository(parser, data_base=self.data_base)
        else:
            bot.error("Please define a unique identifier to add.")

    def get_or_create(self, uid):
        """Determine if a repo exists.
        """
        try:
            repo = self.get(uid, exact=True)
        except:
            repo = self.add(uid)
        return repo

    def get(self, uid=None, exact=False):
        """Get a software repo based on a uid. If exact is not needed, we can
           search for a match based on the partial uid.  If exact is False, 
           and a uid is not provided, get the last repository created.
        """
        if not uid and not exact:
            repos = get_latest_modified(self.data_base, pattern="metadata*.json")
            if repos:
                uid = (
                    repos.replace("metadata.json", "")
                    .replace(self.data_base, "")
                    .strip("/")
                )
            if not uid or not repos:
                raise NoReposError

        parser = get_parser(uid, config=self.config)
        return SoftwareRepository(parser, exists=True, data_base=self.data_base)

    def update(self, repo):
        """Update a repository by retrieving metadata, and then calling update 
           on the software repository to save it.
        """
        data = repo.parser.get_metadata()
        if data:
            repo.update()

    def label(self, repo, key, value, force=False):
        """Update a repository with a specific key/value pair.
        """
        if key in repo.data and not force:
            raise RuntimeError(
                f"{key} is already defined for {repo.uid}. Use --force to overwrite."
            )
        bot.debug(f"Adding key {key}:{value}")
        repo.update({key: value})

    def search(self, query):
        """A filesystem search can only support returning results with filenames
        """
        results = []
        for repo in self.list_repos():
            if re.search(query, repo[0], re.IGNORECASE):
                results.append(repo)
        return results

    def delete_repo(self, uid):
        """delete a repo based on a specific identifier. 
        """
        if self.exists(uid):
            repo = self.get(uid)
            os.remove(repo.filename)
            bot.info(f"{uid} has been removed.")
            return True

        bot.error(f"{uid} does not exist in the database.")
        return False

    def delete_parser(self, name):
        """delete all repos for a parser, based on executor's name (str).
        """
        parser_dir = os.path.join(self.data_base, name)
        if not os.path.exists(parser_dir):
            bot.info(f"Executor {parser_dir} directory does not exist.")
            return False
        shutil.rmtree(parser_dir)
        return True

    def iter_parsers(self, fullpath=False):
        """list executors based on the subfolders in the base database folder.
        """
        for contender in os.listdir(self.data_base):
            contender = os.path.join(self.data_base, contender)
            if os.path.isdir(contender):
                if not fullpath:
                    yield os.path.basename(contender)
                else:
                    yield contender

    def list_repos(self, name=None):
        """list software repositories, either under a particular parser name
           or just under all parsers. This returns repos in rows to be printed
           (or otherwise parsed).
        """
        listpath = self.data_base
        if name:
            listpath = os.path.join(listpath, name)
        rows = []
        for filename in recursive_find(listpath, pattern="metadata*.json"):
            rows.append(
                [
                    filename.replace("metadata.json", "")
                    .replace(self.data_base, "")
                    .strip("/")
                ]
            )
        return rows


class SoftwareRepository:
    """A software repository is a filesystem representation of a repo. It can 
       take a uid, determine if the repo exists, and then interact with the 
       metadata for it. If the repo is instantiated without a unique id
       it is assumed to not exist yet, otherwise it must already
       exist.
    """

    def __init__(self, parser, data_base, exists=False):
        """A SoftwareRepository holds some uid for a parser, and controls
           interaction with the filesystem. 

           Arguments:
             parser (str)    : the parser
             data_base (str) : the path where the database exists.
             exists (bool)   : if True, must already exists (default is False)
        """
        self.uid = parser.uid
        self.parser = parser
        self.data_base = data_base
        self.data = {}
        self.criteria = {}
        self.taxonomy = {}
        self.create(exists)

    @property
    def url(self):
        return self.parser.get_url(self.data.get("data", ""))

    @property
    def avatar(self):
        return self.parser.get_avatar(self.data.get("data", ""))

    @property
    def description(self):
        return self.parser.get_description(self.data.get("data", ""))

    @property
    def filename(self):
        return os.path.join(self.parser_dir, "metadata.json")

    @property
    def parser_dir(self):
        return os.path.join(self.data_base, self.parser.uid)

    def update(self, updates=None):
        """Update a data file. This means reading, updating, and writing.
        """
        updates = updates or {}
        self.data.update(updates)
        self.save()

    def update_criteria(self, uid, username, response):
        """Update a criteria, meaning adding a True/False answer to the 
           unique id for the user. We are currently assuming that criteria
           have yes/no responses, and True == yes, False == no.
        """
        if uid not in self.criteria:
            self.criteria[uid] = {}
        if response:
            self.criteria[uid][username] = response

    def create(self, should_exist=False):
        """create the filename if it doesn't exist, otherwise if it should (and
           does not) exit on error.
        """
        if should_exist:
            if not os.path.exists(self.filename):

                # Might be provided prefix
                contenders = glob("%s*" % os.path.join(self.data_base, self.parser.uid))
                if len(contenders) == 1:
                    self.parser.uid = re.sub(
                        "(%s/|[.]json)" % self.data_base, "", contenders[0],
                    )

                elif len(contenders) > 1:
                    raise MultipleReposExistError(self.parser.uid)
                else:
                    raise RepoNotFoundError(self.parser.uid)
            self.data = self.load()
            self.taxonomy = self.load_taxonomy()
            self.criteria = self.load_criteria()

        if not os.path.exists(self.parser_dir):
            mkdir_p(self.parser_dir)

        # If it's the first time saving, create basic file
        if not should_exist:
            self.data = {
                "parser": self.parser.name,
                "uid": self.parser.uid,
                "url": self.parser.get_url(),
                "data": self.parser.export(),
            }
            self.save()

    def export(self):
        """wrapper to expose the executor.export function
        """
        return self.parser.export()

    def save(self):
        """Save a json object (metadata.json) for the software repository.
        """
        write_json(self.data, self.filename)

    def summary(self):
        return self.parser.summary()

    def load(self):
        """Given a software uid, load data from filename.
        """
        if os.path.exists(self.filename):
            return read_json(self.filename)

    def get_criteria(self):
        """Get loaded criteria"""
        return self.criteria

    def get_taxonomy(self):
        """Get loaded taxonomy"""
        return self.taxonomy

    def load_criteria(self):
        """Given a repository directory, load criteria files if they exist
        """
        criteria = {}
        for filename in glob(f"{self.parser_dir}/criteria*.tsv"):
            uid = (
                os.path.basename(filename).replace("criteria-", "").replace(".tsv", "")
            )
            content = read_file(filename)
            if uid not in criteria:
                criteria[uid] = {}
            for row in content:
                username, response = row.split("\t")
                criteria[uid][username] = response
        return criteria

    def save_criteria(self):
        """Save criteria to file. Each file is named based on the criteria id,
           and is a tab separated file that includes the username and response. 
        """
        for uid, responses in self.criteria.items():
            filename = os.path.join(self.parser_dir, "criteria-%s.tsv" % uid)
            # Sort based on username
            rows = ["%s\t%s" % (k, v) for k, v in sorted(responses.items())]
            write_file(filename, "\n".join(rows))
            bot.debug(f"{uid} saved to {filename}")

    def load_taxonomy(self):
        """Given a repository directory, load taxonomy annotations if they exist
           The taxonomy.tsv file should be a tab separated file with:
           username	category-unique-id. This means that we keep a record of
           who has categorized what, and load this information into the
           taxonomy dictionary (organized by the category-unique-id which
           then has a total count and list of users).
        """
        taxonomy = {}
        taxonomy_file = os.path.join(self.parser_dir, "taxonomy.tsv")
        if os.path.exists(taxonomy_file):
            content = read_file(taxonomy_file)
            for row in content:
                username, uids = row.split("\t")
                taxonomy[username] = [x.strip() for x in uids.split(",")]
        return taxonomy

    def save_taxonomy(self):
        """Save taxonomy to file. Each file is named taxonomy.tsv,
           and is a tab separated file that includes the username and response. 
        """
        filename = os.path.join(self.parser_dir, "taxonomy.tsv")
        rows = ["%s\t%s" % (k, ",".join(v)) for k, v in sorted(self.taxonomy.items())]
        write_file(filename, "\n".join(rows))
        bot.debug(f"{self.uid} saved to {filename}")

    # Annotation

    def has_criteria_annotation(self, uid, username):
        """Determine if a repository has been annotated by a user.
        """
        if uid not in self.criteria:
            return False
        if username not in self.criteria[uid]:
            return False
        return True

    def has_taxonomy_annotation(self, username):
        """Determine if a repository has been annotated by a user.
        """
        if username not in self.taxonomy:
            return False
        return True



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from rse.main import Encyclopedia


def main(args, extra):

    # Clear a parser, uid, or target
    enc = Encyclopedia(config_file=args.config_file, database=args.database)
    enc.clear(args.target, noprompt=args.force)



import os
from .message import bot

# Shared logging import for both client and default.py (for headless)
RSE_LOG_LEVEL = os.environ.get("RSE_LOG_LEVEL", "INFO")
RSE_LOG_LEVELS = ["DEBUG", "CRITICAL", "ERROR", "WARNING", "INFO", "QUIET", "FATAL"]
if RSE_LOG_LEVEL not in RSE_LOG_LEVELS:
    RSE_LOG_LEVEL = "INFO"



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from rse.exceptions import (
    MissingDatabaseString,
    NoReposError,
    MultipleReposExistError,
    RepoNotFoundError,
)
from rse.main.database.base import Database
from rse.main.parsers import get_parser
from rse.main.parsers.base import ParserBase

from sqlalchemy import create_engine, desc
from sqlalchemy.orm import scoped_session, sessionmaker
from sqlalchemy import or_

import logging
import json

bot = logging.getLogger("rse.main.database.relational")


class RelationalDatabase(Database):
    """A RelationalDatabase is a more robust relational datbase (to sqlite).
       Since the global database property can be any of postgresql, mysql+pysq;,
       it is defined on init. The sqlite database also uses this class, but defines
       a custom init function to handle the rse.db file.
    """

    def __init__(self, config_dir, config=None, **kwargs):
        """init for the filesystem ensures that the base folder (named 
           according to the studyid) exists.
        """
        self.database = kwargs.get("database")
        self.config = config
        database_string = kwargs.get("database_string")
        if not database_string:
            raise MissingDatabaseString

        # The database url includes the type and string
        self.db = "%s://%s" % (self.database, database_string)
        self.create_database()

    def create_database(self):
        """create the databsae based on the string, whether it's relational or
           sqlite. self.db must be defined.
        """
        from rse.main.database.models import Base

        self.engine = create_engine(self.db)
        self.session = scoped_session(
            sessionmaker(autocommit=False, autoflush=False, bind=self.engine)
        )
        Base.query = self.session.query_property()
        Base.metadata.create_all(bind=self.engine)
        self.Base = Base

    # Global

    def exists(self, uid):
        """Determine if a repo exists.
        """
        from rse.main.database.models import SoftwareRepository

        parser = get_parser(uid, config=self.config)
        repo = SoftwareRepository.query.filter(
            SoftwareRepository.uid == parser.uid
        ).first()
        return repo is not None

    def get_or_create(self, uid):
        """Determine if a repo exists.
        """
        from rse.main.database.models import SoftwareRepository

        parser = get_parser(uid, config=self.config)
        repo = SoftwareRepository.query.filter(
            SoftwareRepository.uid == parser.uid
        ).first()
        if not repo:
            repo = self.add(uid)
        return repo

    def clear(self):
        """clear (delete) all repos. This could be improved to cascade instead.
        """
        from rse.main.database.models import SoftwareRepository

        SoftwareRepository.query.delete()
        self.session.commit()
        return True

    # Add or Update requires executor

    def add(self, uid):
        """Create a new repo based on a uid that matches to a parser.
        """
        from rse.main.database.models import SoftwareRepository

        parser = get_parser(uid, config=self.config)
        if not self.exists(parser.uid):
            data = parser.get_metadata()

            # If it's a parser handoff
            if isinstance(data, ParserBase):
                parser = data
                data = parser.data

            if data:
                repo = SoftwareRepository(
                    uid=parser.uid, parser=parser.name, data=json.dumps(parser.export())
                )
                self.session.add(repo)
                self.session.commit()
                bot.info(f"{parser.uid} was added to the the database.")
                repo.parser = parser
                return repo

    def update(self, repo, updates=None):
        """update a repo with a json dictionary.
        """
        # Return of None indicates non-success
        data = repo.parser.get_metadata()
        if data:
            updates = updates or {}
            updates.update(repo.parser.export())

            # Load the previous data to update
            data = {}
            if repo.data:
                data = json.loads(repo.data)
            data.update(updates)
            repo.data = json.dumps(data)
            self.session.add(repo)
            self.session.commit()
            return repo

    def label(self, repo, key, value, force=False):
        """Update a repository with a specific key/value pair.
        """
        data = {}
        if repo.data:
            data = json.loads(repo.data)

        if key in data and not force:
            raise RuntimeError(
                f"{key} is already defined for {repo.uid}. Use --force to overwrite."
            )
        data.update({key: value})
        bot.debug(f"Adding key {key}:{value}")
        repo.data = json.dumps(data)
        self.session.add(repo)
        self.session.commit()
        return repo

    # Get, delete, etc. only require uid

    def get(self, uid=None):
        """Get a repo based on a uid. Exits on error if doesn't exist. If
           a uid is not provided, get the last updated repository.
        """
        from rse.main.database.models import SoftwareRepository

        # Retrieve either the last repo, or the one with a specific uid
        if not uid:
            repo = (
                self.session.query(SoftwareRepository)
                .order_by(desc("timestamp"))
                .first()
            )
            parser = get_parser(repo.uid, config=self.config)
            if not repo:
                raise NoReposError
        else:
            parser = get_parser(uid, config=self.config)
            repo = SoftwareRepository.query.filter(
                SoftwareRepository.uid == parser.uid
            ).first()

            # If an exact match isn't there, look for partial match
            if not repo:
                query = "%" + parser.uid + "%"
                query = self.session.query(SoftwareRepository).filter(
                    SoftwareRepository.uid.ilike(query)
                )
                results = self.session.execute(query).fetchall()
                if len(results) == 1:
                    return self.get(results[0][0])
                elif len(results) > 1:
                    raise MultipleReposExistError(parser.uid)
                else:
                    raise RepoNotFoundError(parser.uid)

        repo.parser = parser
        return repo

    def delete_repo(self, uid):
        """delete a repo based on a specific repo id.
        """
        from rse.main.database.models import SoftwareRepository

        repo = self.get(uid)
        if not repo:
            bot.error(f"{uid} does not exist in the database.")
            return False
        SoftwareRepository.query.filter(SoftwareRepository.uid == repo.uid).delete()
        self.session.commit()
        bot.info(f"{uid} has been removed.")
        return True

    def delete_parser(self, name):
        """delete all repos for a parser, based on parser's name (str).
        """
        from rse.main.database.models import SoftwareRepository

        deleted_items = False
        for repo in SoftwareRepository.query.filter(
            SoftwareRepository.parser_name == name
        ):
            deleted_items = True
            self.session.delete(repo)
        self.session.commit()
        return deleted_items

    def list_repos(self, name=None):
        """list repos, either under a particular parser name (if provided)
           or just the parsers. 
        """
        from rse.main.database.models import SoftwareRepository

        if name:
            repos = SoftwareRepository.query.filter(
                SoftwareRepository.parser_name == name
            )
        else:
            repos = SoftwareRepository.query.all()

        rows = []
        for repo in repos:
            rows.append([repo.uid])
        return rows

    def search(self, query):
        """Search across the database for a particular query.
        """
        from rse.main.database.models import SoftwareRepository

        # Ensure that query can be part of a larger string
        query = "%" + query + "%"

        query = self.session.query(SoftwareRepository).filter(
            or_(
                SoftwareRepository.data.ilike(query),
                SoftwareRepository.uid.ilike(query),
            )
        )
        # list of tuples, (uid, datetime, executor]
        results = self.session.execute(query).fetchall()
        return [[r[0], str(r[2]), str(r[1])] for r in results]



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from rse.defaults import RSE_API_ENDPOINT
import requests
import logging
import sys

bot = logging.getLogger("rse.main.taxonomy")
parser_regex = "github"


def get_taxonomy():
    """Get taxonomy from the default endpoint.
    """
    response = requests.get(f"{RSE_API_ENDPOINT}/taxonomy/")
    if response.status_code != 200:
        sys.exit(f"Problem with getting {RSE_API_ENDPOINT}/taxonomy/")
    taxonomy = response.json()["data"]
    # Sort by the ordering
    return sorted(taxonomy, key=lambda i: i["path"])



#!/usr/bin/env python
"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

import os
import sys
import pytest


def test_parser_github(tmp_path):
    """Test the github parser.
    """
    from rse.main.parsers import GitHubParser

    for repo in [
        "singularityhub/sregistry",
        "github.com/singularityhub/sregistry",
        "git@github.com:singularityhub/sregistry.git",
    ]:
        parser = GitHubParser(repo)
        assert parser.uid == "github/singularityhub/sregistry"
        assert parser.summary()

    # Only test one get of data
    assert parser.get_metadata()
    data = parser.export()
    for key in ["timestamp", "url", "html_url"]:
        assert key in data



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from .biotools import BioToolsScraper
from .hal import HalScraper
from .joss import JossScraper
from .rsnl import RSNLScraper
import re


def get_named_scraper(name, config=None):
    """get a named scraper, meaning determining based on name
    """
    scraper = None
    if re.search("(joss|journal of open source software)", name, re.IGNORECASE):
        scraper = JossScraper()
    elif re.search("(biotool|bio[.]tool)", name, re.IGNORECASE):
        scraper = BioToolsScraper()
    elif re.search("hal", name, re.IGNORECASE):
        scraper = HalScraper()
    elif re.search("(researchsoftwarenl|rsnl)", name, re.IGNORECASE):
        scraper = RSNLScraper()

    if not scraper:
        raise NotImplementedError(f"There is no matching scraper for {name}")

    scraper.config = config
    return scraper



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

import logging
import requests
from rse.utils.urls import check_response

from .base import ParserBase

bot = logging.getLogger("rse.main.parsers.github")


class GitHubParser(ParserBase):

    name = "github"
    matchstring = "github"

    def __init__(self, uid=None, **kwargs):
        super().__init__(uid)

    def _set_uid(self, uid):
        """Given some kind of GitHub url, parse the uid
        """
        uid = uid.replace(":", "/")
        owner, repo = uid.replace(".git", "").split("/")[-2:]
        return "{}/{}".format(owner, repo)

    def load_secrets(self):
        """load secrets, namely the GitHub token
        """
        self.token = self.get_setting("TOKEN")

    def get_url(self, data=None):
        """a common function for a parser to return the html url for the
           upper level of metadata
        """
        data = data or self.data
        return data.get("html_url")

    def get_avatar(self, data=None):
        """a common function for a parser to return an image.
        """
        data = data or self.data
        return data.get("owner", {}).get("avatar_url", "")

    def get_description(self, data=None):
        """a common function for a parser to return a description.
        """
        data = data or self.data
        return data.get("description")

    def get_metadata(self, uri=None):
        """Retrieve repository metadata. The common metadata (timestamp) is
           added by the software repository parser, and here we need to
           ensure that the url field is populated with a correct url.

           Arguments:
           uri (str) : a repository uri string to override one currently set
        """
        if uri:
            self.set_uri(uri)
        self.load_secrets()
        repo = "/".join(self.uid.split("/")[-2:])
        url = "https://api.github.com/repos/%s" % (repo)
        headers = {
            "Accept": "application/vnd.github.symmetra-preview+json",
        }
        if self.token:
            headers["Authorization"] = "token %s" % self.token

        response = requests.get(url, headers=headers)

        # Successful query!
        self.data = check_response(response)
        return self.data



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from rse.main import Encyclopedia
import sys


def main(args, extra):

    # Clear a parser, uid, or target
    enc = Encyclopedia(config_file=args.config_file, database=args.database)

    # Pass the encyclopedia object to start a server
    try:
        from rse.app.server import start

        start(
            port=args.port,
            client=enc,
            debug=args.debug,
            level=args.log_level,
            disable_annotate=args.disable_annotate,
        )
    except:
        sys.exit(
            "You must 'pip install rse[app]' 'pip install rse[all]' to use the dashboard."
        )



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from rse.utils.urls import get_user_agent
from rse.main.parsers import get_parser
import logging
import requests
import random
import sys
import re
from time import sleep

from .base import ScraperBase

bot = logging.getLogger("rse.main.scrapers.joss")


class JossScraper(ScraperBase):

    name = "joss"

    def __init__(self, query=None, **kwargs):
        super().__init__(query)

    def latest(self, paginate=False):
        """The scraper should expose a function to populate self.results with
           some number of latest entries. Unlike a search, a latest scraper does
           not by default paginate. The user needs to interact directly with
           the Python client to do a scrape for all papers in JoSS.
        """
        url = "https://joss.theoj.org/papers/published.atom"
        return self.scrape(url, paginate=paginate)

    def search(self, query, paginate=True):
        """The scraper should expose a function to populate self.results with
           a listing based on matching a search criteria.
        """
        url = "https://joss.theoj.org/papers/search?q=%s" % query
        return self.scrape(url, paginate=paginate)

    def scrape(self, url, paginate=False):
        """A shared function to scrape a set of repositories. Since the JoSS
           pages for a search and the base are the same, we can use a shared
           function.
        """
        try:
            from bs4 import BeautifulSoup
        except ImportError:
            sys.exit("BeautifulSoup is required. pip install rse[scraper].")

        # Handle pagination
        while url is not None:

            response = requests.get(url, headers={"User-Agent": get_user_agent()})
            soup = BeautifulSoup(response.text, "html.parser")
            url = None
            for link in soup.find_all("link", href=True):

                # Sleep for a random amount of time to give a rest!
                sleep(random.choice(range(1, 10)) * 0.1)
                paper_url = link.attrs.get("href", "")

                # If we don't have the next page yet
                if link.attrs.get("rel") is not None and url is None and paginate:
                    if link.attrs.get("rel")[0] == "next":
                        url = link.attrs.get("href")

                # Retrieve page with paper metadata that we need
                if re.search(
                    "https://joss.theoj.org/papers/10.[0-9]{5}/joss.[0-9]{5}", paper_url
                ):
                    response = requests.get(
                        paper_url, headers={"User-Agent": get_user_agent()}
                    )
                    paper_soup = BeautifulSoup(response.text, "html5lib")

                    # Find links that we need
                    repo = {}
                    for link in paper_soup.find_all("a", href=True):
                        if "Software repository" in link.text:
                            repo["url"] = link.attrs.get("href", "")
                        elif "Software archive" in link.text:
                            repo["doi"] = link.attrs.get("href", "")

                    if repo.get("url") and repo.get("doi"):
                        bot.info("Found repository: %s" % repo["url"])
                        self.results.append(repo)

        return self.results

    def create(self, database=None, config_file=None):
        """After a scrape (whether we obtain latest or a search query) we
           run create to create software repositories based on results.
        """
        from rse.main import Encyclopedia

        client = Encyclopedia(config_file=config_file, database=database)
        for result in self.results:
            uid = result["url"].split("//")[-1]
            repo = get_parser(uid)

            # Add results that don't exist
            if not client.exists(repo.uid):
                client.add(repo.uid)
                client.label(repo.uid, key="doi", value=result.get("doi"))



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from rse.main import Encyclopedia
from rse.exceptions import RepoNotFoundError, NoReposError
import json
import logging

bot = logging.getLogger("rse.client")


def main(args, extra):

    # Create a research software encyclopedia
    enc = Encyclopedia(config_file=args.config_file, database=args.database)

    try:
        repo = enc.get(args.uid)
        print(json.dumps(repo.load(), indent=4))
    except NoReposError:
        bot.error(
            "There are no software repositories in the database! Use rse add to add some."
        )
    except RepoNotFoundError:
        bot.error(f"{args.uid} does not exist in the database. Use rse add to add it.")



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

import json

from rse.main import Encyclopedia
from rse.logger.message import bot


def analyze(args, extra):
    """analyze is intended to provide calculations for repos, by default including
       all taxonomy categories and criteria. For a custom set, the user should
       interact with the client directly.
    """
    client = Encyclopedia(config_file=args.config_file, database=args.database)
    result = client.analyze(repo=args.repo, cthresh=args.cthresh, tthresh=args.tthresh)

    # Create rows for criteria and taxonomy
    clookup = {c["uid"]: c for c in client.list_criteria()}
    tlookup = {t["uid"]: t for t in client.list_taxonomy()}
    crows = [
        [response, clookup[uid]["name"]] for uid, response in result["criteria"].items()
    ]
    trows = [
        [str(count), tlookup[uid]["name"]] for uid, count in result["taxonomy"].items()
    ]

    bot.info(f"Summary for {result['repo']}")
    bot.info("\nCriteria")
    bot.table(crows)
    bot.info("\nTaxonomy")
    bot.table(trows)


def summary(args, extra):
    """summary will return general counts for taxonomy, criteria, and users"""

    client = Encyclopedia(config_file=args.config_file, database=args.database)
    metrics = client.summary(args.repo)

    # Export only taxonomy metrics
    if args.metric_type == "taxonomy":
        updated = {}
        for field in ["taxonomy", "taxonomy-count", "repos"]:
            updated[field] = metrics[field]
        metrics = updated

    # Export only criteria metrics
    elif args.metric_type == "criteria":
        updated = {}
        for field in ["criteria", "criteria-count", "repos"]:
            updated[field] = metrics[field]
        metrics = updated

    # Export only user metrics
    elif args.metric_type == "users":
        updated = {}
        for field in ["users-count", "users", "repos"]:
            updated[field] = metrics[field]
        metrics = updated

    # Print metrics as json or table
    if metrics:
        print(json.dumps(metrics, indent=4))



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

__version__ = "0.0.21"
AUTHOR = "Vanessa Sochat"
AUTHOR_EMAIL = "vsochat@stanford.edu"
NAME = "rse"
PACKAGE_URL = "https://github.com/rseng/rse"
KEYWORDS = "rse,software,research software,database,taxonomy"
DESCRIPTION = "tools for assessment and categorization of research software"
LICENSE = "LICENSE"

################################################################################
# Global requirements


INSTALL_REQUIRES = (("requests", {"min_version": "2.23.0"}),)

APP_REQUIRES = (
    ("Flask", {"min_version": "1.0.2"}),
    ("Flask-SocketIO", {"min_version": "3.1.2"}),
    ("flask-restful", {"min_version": "0.3.8"}),
    ("gevent-websocket", {"min_version": "0.10.1"}),
)

SCRAPER_REQUIRES = (
    ("beautifulsoup4", {"min_version": "4.9.0"}),
    ("html5lib", {"min_version": "1.0.1"}),
)
DATABASE_REQUIRES = (("sqlalchemy", {"min_version": None}),)
TESTS_REQUIRES = (("pytest", {"min_version": "4.6.2"}),)


ALL_REQUIRES = INSTALL_REQUIRES + DATABASE_REQUIRES + APP_REQUIRES + SCRAPER_REQUIRES



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from rse.main.config import Config
from rse.defaults import RSE_DATABASE, RSE_PARSERS, RSE_CONFIG_FILE

from rse.exceptions import RepoNotFoundError, RepoMetadataExistError
from rse.main.database import init_db
from rse.utils.prompt import confirm, choice_prompt
from rse.utils.file import read_file
from rse.utils.command import get_github_username
from rse.utils.urls import repository_regex
from rse.main.parsers import get_parser
from rse.main.criteria import get_criteria
from rse.main.taxonomy import get_taxonomy
from rse.logger.message import bot as message

import logging
import os
import re

bot = logging.getLogger("rse.main")
parser_regex = "github"


class Encyclopedia:
    """An encyclopedia is one or more namespaces to store research
       software. By default, we create a structure on the filesystem,
       however an sqlite database (or other) can be used.
    """

    def __init__(self, config_file=None, database=None, generate=False):
        """create a software repository. We take a config file, which should
           sit at the root of the repository, and then parse the subfolders
           accordingly.
        """
        self.config = Config(config_file or RSE_CONFIG_FILE, generate=generate)
        self.config_dir = os.path.dirname(self.config.configfile)
        self.initdb(database)

    def initdb(self, database):
        """setup the rse home (where the config directory is stored) and the
           database specification. If a database string is required (and not
           provided) alert the user and exit on error).

           Arguments:
            - config_dir (str) : the configuration directory (home for rse)
            - database (str) : a string to specify the database setup
        """
        self.database = (
            database
            or RSE_DATABASE
            or self.config.get("DEFAULT", "database")
            or "filesystem"
        )
        database_string = self.config.get("DEFAULT", "databaseconnect")
        bot.info("Database: %s" % self.database)

        # Supported database options
        valid = ("sqlite", "postgresql", "mysql+pymysql", "filesystem")
        if not self.database.startswith(valid):
            bot.warning(
                "%s is not yet a supported type, saving to filesystem." % self.database
            )
            self.database = "filesystem"

        # Create database client with functions for database type
        self.db = init_db(
            self.database,
            config_dir=self.config_dir,
            database_string=database_string,
            config=self.config,
        )

    def exists(self, uid):
        """based on a parser type and unique identifier, determine if software
           exists in the database
        """
        parser = get_parser(uid, config=self.config)
        return self.db.exists(parser.uid)

    def list(self, name=None):
        """A wrapper to the database list_repos function. Optionally take
           a whole parser name (e.g., github) or just a specific uid. No
           parser indicates that we list everything.
        """
        return self.db.list_repos(name)

    def list_criteria(self):
        """Get a listing of criteria from the rse API
        """
        if not hasattr(self, "criteria"):
            self.criteria = get_criteria()
        return self.criteria

    def list_taxonomy(self):
        """Get a listing of a flattened taxonomy from the rse API
        """
        if not hasattr(self, "taxonomy"):
            self.taxonomy = get_taxonomy()
        return self.taxonomy

    def bulk_add(self, filename):
        """Given a filename with a single list of repos, add each
        """
        repos = []
        if os.path.exists(filename):
            for name in read_file(filename):
                uid = name.strip()
                repos += [self.add(uid, quiet=True)] or []
        return repos

    def bulk_update(self, filename):
        """Given a filename with a single list of repos, add each
        """
        repos = []
        if os.path.exists(filename):
            for name in read_file(filename):
                uid = name.strip()
                try:
                    repos += [self.update(uid)]
                except RepoNotFoundError:
                    pass
        return repos

    def add(self, uid, quiet=False):
        """A wrapper to add a repository to the software database.
        """
        if not self.exists(uid):
            repo = self.db.add(uid)
            return repo
        if not quiet:
            bot.error(f"{uid} already exists in the database.")

    def get(self, uid=None):
        """A wrapper to get a repo id from the database. If an id is not provided,
           will return the last updated repo based on timestamp of file or database.
        """
        return self.db.get(uid)

    def get_or_create(self, uid):
        return self.db.get_or_create(uid)

    def clear(self, target=None, noprompt=False):
        """clear takes a target, and that can be a uid, parser, or none
           We ask the user for confirmation.
        """
        # Case 1: no target indicates clearing all
        if not target:
            if noprompt or confirm(
                "This will delete all software in the database, are you sure?"
            ):
                return self.db.clear()

        # Case 2: it's a parser
        elif target in RSE_PARSERS:
            if noprompt or confirm(
                f"This will delete all {target} software in the database, are you sure?"
            ):
                return self.db.delete_parser(target)

        # Case 3, it's a specific software identifier
        elif re.search(parser_regex, target):
            if noprompt or confirm(
                f"This will delete software {target}, are you sure?"
            ):
                return self.db.delete_repo(target)

        else:
            raise RuntimeError(f"Unrecognized {target} to clear")

    def update(self, uid):
        """Update an existing software repository.
        """
        try:
            repo = self.get(uid)
            self.db.update(repo)
            bot.info(f"{repo.uid} has been updated.")
            return repo
        except RepoNotFoundError:
            bot.error(f"{uid} does not exist.")

    def label(self, uid, key, value, force=False):
        """Update an existing software repository with a specific label.
        """
        try:
            repo = self.get(uid)
            self.db.label(repo, key, value, force=force)
            bot.info(f"{repo.uid} has been updated.")
            return repo
        except RepoMetadataExistError:
            bot.error(
                f"{repo.uid} already has value for {key}. Use --force to overwrite."
            )
        except RepoNotFoundError:
            bot.error(f"{uid} does not exist.")

    def search(self, query):
        """Search across commands and general metadata for a string of interest.
           We use regular expressions (re.search) so they are supported.
           Search is only available for non-filesystem databases.
        """
        results = self.db.search(query)
        if results:
            return results
        bot.info(f"No results matching {query}")

    # Save Handlers

    def save_criteria(self, repo):
        """Given a repository that can be a handle to a filesystem entry
           or database, save the criteria
        """
        # The filesystem database saves at the end
        if hasattr(repo, "save_criteria"):
            repo.save_criteria()

        # Relational saves the database item
        else:
            self.db.update(repo)

    def save_taxonomy(self, repo, username, uids):
        """Given a repository that can be a handle to a filesystem entry
           or database, save the criteria
        """
        if hasattr(repo, "save_taxonomy"):
            repo.taxonomy[username] = uids
            repo.save_taxonomy()
        else:
            repo.update_taxonomy(username, uids)
            self.db.update(repo)

    # Metrics
    def analyze_bulk(
        self,
        cthresh=0.5,
        tthresh=1,
        taxonomy_uids=None,
        criteria_uids=None,
        include_empty=False,
    ):
        """analyze takes a repository and calculates a "final answer" based on user provided
           thresholds
        """
        results = []
        for repo in self.list():
            result = self.analyze(
                repo[0],
                cthresh=cthresh,
                tthresh=tthresh,
                taxonomy_uids=taxonomy_uids,
                criteria_uids=criteria_uids,
            )
            if not result["taxonomy"] and not result["criteria"] and not include_empty:
                continue
            results.append(result)
        return results

    def analyze(
        self, repo, cthresh=0.5, tthresh=1, taxonomy_uids=None, criteria_uids=None
    ):
        """analyze takes a repository and calculates a "final answer" based on user provided
           thresholds
        """
        # If taxonomy or criteria lists aren't defined, use all
        if not taxonomy_uids:
            taxonomy_uids = [x["uid"] for x in self.list_taxonomy()]
        if not criteria_uids:
            criteria_uids = [x["uid"] for x in self.list_criteria()]

        parser = get_parser(repo, config=self.config)
        repo = self.get(parser.uid)
        metrics = {"repo": parser.uid, "criteria": {}, "taxonomy": {}}

        # Calculate "final" answers for each criteria based on votes and threshold
        counts = {}
        for name, votes in repo.get_criteria().items():
            # Skip criteria if not important
            if name not in criteria_uids:
                continue
            if name not in counts:
                counts[name] = {"yes": 0, "no": 0, "total": 0}
            for username, response in votes.items():
                counts[name][response] += 1
                counts[name]["total"] += 1

        # Calculate final answers!
        for name, summary in counts.items():
            if summary["yes"] / summary["total"] >= cthresh:
                metrics["criteria"][name] = "yes"
            else:
                metrics["criteria"][name] = "no"

        counts = {}
        for username, categories in repo.get_taxonomy().items():
            for category in categories:
                if category not in counts:
                    counts[category] = 0
                counts[category] += 1

        # Include those above the requested threshold
        for name, count in counts.items():
            if count >= tthresh:
                metrics["taxonomy"][name] = count

        return metrics

    def summary(self, repo=None):
        """Summarize metrics for the entire database if uid is not defined,
           or one specific repository.
        """
        if repo is None:
            repos = self.list()
            metrics = {"repos": len(repos)}
        else:
            parser = get_parser(repo, config=self.config)
            repos = [[parser.uid]]
            metrics = {"repo": parser.uid}

        # Add taxonomy and criteria items
        metrics["taxonomy-count"] = len(self.list_taxonomy())
        metrics["criteria-count"] = len(self.list_criteria())
        metrics["users"] = {}
        metrics["taxonomy"] = {}
        metrics["criteria"] = {}

        # Count annotations for
        for repo in repos:
            parser = get_parser(repo[0], config=self.config)
            repo = self.get(parser.uid)

            if not repo.criteria and not repo.taxonomy:
                continue

            # Add repository to summary metrics
            metrics["taxonomy"][repo.uid] = {}
            metrics["criteria"][repo.uid] = {}

            # Derive all users that have annotated taxonomy/criteria
            users = set()
            for name, votes in repo.get_criteria().items():
                [users.add(user) for user in votes.keys()]
                if name not in metrics["criteria"][repo.uid]:
                    metrics["criteria"][repo.uid] = {"yes": 0, "no": 0}
                for vote in votes.values():
                    metrics["criteria"][repo.uid][vote] += 1

            # Update criteria annotations
            for user in users:
                if user not in metrics["users"]:
                    metrics["users"][user] = {
                        "criteria-annotations": 0,
                        "taxonomy-annotations": 0,
                    }
                metrics["users"][user]["criteria-annotations"] += 1

            # Derive all users that have annotated taxonomy/criteria
            users = set()
            for username, categories in repo.get_taxonomy().items():
                users.add(username)
                for category in categories:
                    if category not in metrics["taxonomy"][repo.uid]:
                        metrics["taxonomy"][repo.uid][category] = 0
                    metrics["taxonomy"][repo.uid][category] += 1

            # Don't add empty entries
            if not repo.taxonomy and repo.uid in metrics["taxonomy"]:
                del metrics["taxonomy"][repo.uid]

            if not repo.criteria and repo.uid in metrics["criteria"]:
                del metrics["criteria"][repo.uid]

        # Add unique users
        metrics["users-count"] = len(metrics["users"])
        return metrics

    # Annotation

    def annotate(self, username, atype, unseen_only=True, repo=None, save=False):
        """Annotate the encyclopedia, either for criteria or taxonomy.
           A username is required for the namespace.
 
           Arguments:
            - username (str) : the user's GitHub username
            - atype (str) : the annotation type
            - unseen_only (bool): annotate only items not seen by username
            - repo (str) : annotate a particular software repository
        """
        # git config user.name
        if not username:
            username = get_github_username()

        if atype == "criteria":
            return self.annotate_criteria(username, unseen_only, repo, save)
        elif atype == "taxonomy":
            return self.annotate_taxonomy(username, unseen_only, repo, save)
        bot.error(f"Unknown annotation type, {atype}.")

    def _import_annotation(self, input_file, username, stop_line="## Criteria"):
        """A general helper (private)  function to import an annotation, meaning
           we parse a repository and return additional lines for parsing.
        """
        if not username or not input_file:
            raise RuntimeError(
                "A username and input file are required to import annotation criteria."
            )

        if not os.path.exists(input_file):
            raise FileNotFoundError(input_file)

        lines = read_file(input_file)
        line = lines.pop(0)

        # Find the repository name
        while stop_line not in line:
            match = re.search(repository_regex, line)
            if match:
                break
            line = lines.pop(0)

        # Retrieve the match
        if not match:
            raise RuntimeError(f"repository pattern not found in {input_file}")
        reponame = match.group()
        parser = get_parser(reponame)
        repo = self.get(parser.uid)
        return repo, lines

    def import_criteria_annotation(self, input_file, username):
        """Given a text file that has a bullet list of (some checked) criteria
           as might be generated in a GitHub issue, read in the file and the
           username to do an annotation. If a user has already done an annotation,
           his or her record is updated.
        """
        repo, lines = self._import_annotation(input_file, username)

        # Now iterate through checklist, update
        for line in lines:
            uid = line.split("criteria-")[-1].strip()
            if "[x]" in line:
                repo.update_criteria(uid, username, "yes")
                print(f"Updating {repo.uid}: {uid}->yes")
            elif re.search("\[]|\[ \]", line):
                repo.update_criteria(uid, username, "no")
                print(f"Updating {repo.uid}: {uid}->no")

        # Save the criteria
        self.save_criteria(repo)

    def import_taxonomy_annotation(self, input_file, username):
        """Given a text file that has a bullet list of (some checked) criteria
           as might be generated in a GitHub issue, read in the file and the
           username to do an annotation. If a user has already done an annotation,
           his or her record is updated.
        """
        repo, lines = self._import_annotation(
            input_file, username, stop_line="## Taxonomy"
        )

        # Now iterate through checklist, update
        uids = []
        for line in lines:
            if "RSE-taxonomy" not in line or "[x]" not in line:
                continue
            uid = line.split("]")[-1].strip()
            if uid.startswith("RSE-taxonomy") and uid not in uids:
                print(f"{repo.uid} adding {uid}")
                uids.append(uid)
        self.save_taxonomy(repo, username, uids)

    def yield_criteria_annotation_repos(self, username, unseen_only=True, repo=None):
        """Given a username, repository, and preference for seen / unseen,
           yield a repository to annotate.
        """
        if repo is None:
            repos = self.list()
        else:
            parser = get_parser(repo, config=self.config)
            repos = [[parser.uid]]
            unseen_only = False

        # yield combinations that don't exist yet, repo first to save changes
        for name in repos:
            repo = self.get(name[0])
            for item in self.list_criteria():
                if unseen_only and not repo.has_criteria_annotation(
                    item["uid"], username
                ):
                    yield repo, item
                elif not unseen_only:
                    yield repo, item

    def yield_taxonomy_annotation_repos(self, username, unseen_only=True, repo=None):
        """Given a username, repository, and preference for seen / unseen,
           yield a repository to annotate.
        """
        if repo is None:
            repos = self.list()
        else:
            parser = get_parser(repo, config=self.config)
            repos = [[parser.uid]]
            unseen_only = False

        # yield combinations that don't exist yet, repo first to save changes
        for name in repos:
            repo = self.get(name[0])
            if unseen_only and not repo.has_taxonomy_annotation(username):
                yield repo
            elif not unseen_only:
                yield repo

    def annotate_criteria(self, username, unseen_only=True, repo=None, save=False):
        """Annotate criteria, meaning we iterate over repos and criteria that 
           match the user request, namely to annotate unseen only, or just
           a particular repository. If the repository is specified, unseen_only
           is assumed False.
        """
        annotations = {}
        last = None
        for repo, criteria in self.yield_criteria_annotation_repos(
            username, unseen_only, repo
        ):

            # Only print repository if not seen yet
            if not last or repo.uid != last.uid:

                # If we have a last repo, we need to save progress
                if last is not None and save is True:
                    self.save_criteria(last)

                if last is not None:
                    annotations[last.uid] = last.criteria

                message.info(f"\n{repo.url} [{repo.description}]:")
                last = repo

            response = choice_prompt(
                criteria["name"],
                choices=["y", "Y", "n", "N", "s", "S", "skip"],
                choice_prefix="y/n or s to skip",
            )

            # The user can skip an answer if wanted
            if response in ["s", "S", "skip"]:
                continue

            repo.update_criteria(criteria["uid"], username, response)

        # Save the last repository
        if last is not None and save is True:
            self.save_criteria(last)

        if last is not None:
            annotations[last.uid] = last.criteria
        return annotations

    def annotate_taxonomy(self, username, unseen_only=True, repo=None, save=False):
        """Annotate taxonomy, meaning we iterate over repos and criteria that 
           match the user request, namely to annotate unseen only, or just
           a particular repository. If the repository is specified, unseen_only
           is assumed False.
        """
        annotations = {}

        # Retrieve the full taxonomy
        items = self.list_taxonomy()
        choices = [str(i) for i, _ in enumerate(items)] + ["s", "S", "skip"]
        prefix = "0:%s or s to skip" % (len(items) - 1)

        for repo in self.yield_taxonomy_annotation_repos(username, unseen_only, repo):

            message.info(f"\n{repo.url} [{repo.description}]:")
            print("How would you categorize this software? [enter one or more numbers]")
            for i, t in enumerate(items):
                example = t.get("example", "")
                name = t.get("name", "")
                if name and example:
                    print(f"[{i}] {name} ({example})")
                elif name:
                    print(f"[{i}] {name}")

            response = choice_prompt(
                "Please enter one or more numbers, separated by spaces",
                choices=choices,
                choice_prefix=prefix,
                multiple=True,
            )

            if response in ["s", "S", "skip"]:
                continue

            # Get the unique ids
            uids = [
                items[int(x)]["uid"]
                for x in set(response.split(" "))
                if int(x) < len(items)
            ]

            # Filesystem database we write filename to repository folder
            self.save_taxonomy(repo, username, uids)
            annotations[repo.uid] = repo.taxonomy

        return annotations



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""


def confirm(prompt, response=False):
    """Used to prompt the user for a yes or no response, and returns True/False.
       This means we can use it like:
       if confirm("Would you like to do the thing?"):
          ....
    """
    prompt = "%s [%s]|%s: " % (prompt, "n", "y")
    while True:
        answer = input(prompt)
        if not answer:
            return response
        if answer not in ["y", "Y", "n", "N"]:
            print("Please respond with y or n.")
            continue
        if answer == "y" or answer == "Y":
            return True
        if answer == "n" or answer == "N":
            return False


def choice_prompt(prompt, choices, choice_prefix=None, multiple=False):
    """Ask the user for a prompt, and only return when one of the requested
       options is provided.

       Parameters
       ==========
       prompt: the prompt to ask the user
       choices: a list of choices that are valid.   
       multiple: allow multiple responses (separated by spaces)
    """
    choice = None
    print(prompt)
    get_input = getattr(__builtins__, "raw_input", input)

    if not choice_prefix:
        choice_prefix = "/".join(choices)
    message = "[%s] : " % (choice_prefix)

    while choice not in choices:
        choice = get_input(message).strip()

        # If multiple allowed, add selection to choices if includes all vaid
        if multiple is True:
            contenders = choice.strip().split(" ")
            if all([x in choices for x in contenders]):
                choices.append(choice)
        message = "Please enter a valid option in [%s]" % (choice_prefix)
    return choice



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

import os


class ScraperBase:
    """A scraper base exists get new records from a resource.
    """

    name = "base"

    def __init__(self, query=None):
        """create a scraper. if a query term is provided, do search. Otherwise
           do a query for latest.
        """
        self.query = query
        self.results = []

    def latest(self, paginate=True):
        """The scraper should expose a function to populate self.results with
           some number of latest entries.
        """
        raise NotImplementedError

    def search(self, paginate=True):
        """The scraper should expose a function to populate self.results with
           a listing based on matching a search criteria.
        """
        raise NotImplementedError

    def create(self, uri, **kwargs):
        """After a scrape (whether we obtain latest or a search query) we
           run create to create software repositories based on results.
        """
        raise NotImplementedError

    def get_setting(self, key, default=None):
        """Get a setting, meaning that we first check the environment, then
           the config file, and then (if provided) a default.
        """
        # First preference to environment
        envar = ("RSE_%s_%s" % (self.name, key)).upper()
        envar = os.environ.get(envar)
        if envar is not None:
            return envar

        # Next preference to config setting
        parser = "scraper.%s" % self.name

        # Parsers instantiated separate from database won't have config
        if not hasattr(self, "config"):
            return default
        if parser not in self.config.config:
            return default
        if key in self.config.config[parser]:
            return self.config.get(parser, key)
        return default

    def summary(self):
        return "[scraper][%s][%s]" % (self.name, len(self.results))



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

import os
import locale
import subprocess
import shlex
import shutil
import tempfile
from .file import read_file


def get_github_username():
    """Get the github username of the user from git config.
    """
    command = Command("git config user.name")
    username, _ = command.execute()
    if command.returncode != 0:
        raise FileNotFoundError(
            "Problem running git config user.name. Please provide GitHub username as argument."
        )
    return username[0].replace("\n", "")


class Capturing:
    """capture output from stdout and stderr into capture object.
       This is based off of github.com/vsoch/gridtest but modified
       to write files. The stderr and stdout are set to temporary files at
       the init of the capture, and then they are closed when we exit. This
       means expected usage looks like:

       with Capturing() as capture:
           process = subprocess.Popen(...)
           
       And then the output and error are retrieved from reading the files:
       and exposed as properties to the client:

           capture.out
           capture.err

       And cleanup means deleting these files, if they exist.
    """

    def __enter__(self):
        self.set_stdout()
        self.set_stderr()
        return self

    def set_stdout(self):
        self.stdout = open(tempfile.mkstemp()[1], "w")

    def set_stderr(self):
        self.stderr = open(tempfile.mkstemp()[1], "w")

    def __exit__(self, *args):
        self.stderr.close()
        self.stdout.close()

    @property
    def out(self):
        """Return output stream. Returns empty string if empty or doesn't exist.
           Returns (str) : output stream written to file
        """
        if os.path.exists(self.stdout.name):
            return read_file(self.stdout.name)
        return ""

    @property
    def err(self):
        """Return error stream. Returns empty string if empty or doesn't exist.
           Returns (str) : error stream written to file
        """
        if os.path.exists(self.stderr.name):
            return read_file(self.stderr.name)
        return ""

    def cleanup(self):
        for filename in [self.stdout.name, self.stderr.name]:
            if os.path.exists(filename):
                os.remove(filename)


class Command:
    """Class method to invoke shell commands and retrieve output and error.
       This class is inspired and derived from utils functions in
       https://github.com/vsoch/scif
    """

    def __init__(self, cmd=None):

        cmd = cmd or []
        self.returncode = None
        self.out = []
        self.err = []

        # If a list isn't provided, split it
        if cmd:
            self.set_command(cmd)

    def set_command(self, cmd):
        """parse is called when a new command is provided to ensure we have
           a list. We don't check that the executable is on the path,
           as the initialization might not occur in the runtime environment.
        """
        if not isinstance(cmd, list):
            cmd = shlex.split(cmd)
        self.cmd = cmd

    def execute(self):
        """Execute a system command and return output and error.
        :param cmd: shell command to execute
        :type cmd: str, required
        :return: Output and Error from shell command
        :rtype: two str objects
        """
        # Reset the output and error records
        self.out = []
        self.err = []

        # The executable must be found, return code 1 if not
        executable = shutil.which(self.cmd[0])
        if not executable:
            self.err = ["%s not found." % self.cmd[0]]
            self.returncode = 1
            return (self.out, self.err)

        # remove the original executable
        args = self.cmd[1:]

        # Use updated command with executable and remainder (list)
        cmd = [executable] + args

        # Capturing provides temporary output and error files
        with Capturing() as capture:
            process = subprocess.Popen(
                cmd,
                stdout=capture.stdout,
                stderr=capture.stderr,
                universal_newlines=True,
            )
            returncode = process.poll()

            # Iterate through the output
            while returncode is None:
                returncode = process.poll()

        # Get the remainder of lines, add return code
        self.out += ["%s\n" % x for x in self.decode(capture.out) if x]
        self.err += ["%s\n" % x for x in self.decode(capture.err) if x]

        # Cleanup capture files and save final return code
        capture.cleanup()
        self.returncode = returncode

        return (self.out, self.err)

    def returnCode(self):
        """Returns the return code from shell command
        :rtype: int
        """

        return self.returncode

    def decode(self, line):
        """Given a line of output (error or regular) decode using the
           system default, if appropriate
        """
        loc = locale.getdefaultlocale()[1]

        try:
            line = line.decode(loc)
        except:
            pass
        return line

    def get_output(self):
        """Returns the output from shell command
        :rtype: str
        """
        return self.out

    def get_error(self):
        """Returns the error from shell command
        :rtype: str
        """
        return self.err



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

import errno
import fnmatch
import json
import os
import pickle
import tempfile


def get_latest_modified(base, pattern="*.json"):
    """Given a folder, get the latest modified file
    """
    files = list(recursive_find(base, pattern))
    if not files:
        return None
    return max(files, key=os.path.getctime)


def recursive_find(base, pattern="*.py"):
    """recursive find will yield python files in all directory levels
       below a base path.

       Arguments:
         - base (str) : the base directory to search
         - pattern: a pattern to match, defaults to *.py
    """
    for root, _, filenames in os.walk(base):
        for filename in fnmatch.filter(filenames, pattern):
            yield os.path.join(root, filename)


def read_file(filename, readlines=True):
    """write_file will open a file, "filename" and write content
       and properly close the file.

       Arguments:
         - filename (str) : the filename to read
         - readlines (bool) : read lines of the file (vs all raw)
    """
    with open(filename, "r") as filey:
        if readlines is True:
            content = filey.readlines()
        else:
            content = filey.read()
    return content


def write_file(filename, content):
    """Write some text content to a file
    """
    with open(filename, "w") as fd:
        fd.write(content)


def write_json(json_obj, filename, pretty=True):
    """write_json will write a json object to file, pretty printed

       Arguents:
        - json_obj (dict) : the dict to print to json
        - filename (str) : the output file to write to
    """
    with open(filename, "w") as filey:
        if pretty:
            filey.writelines(json.dumps(json_obj, indent=4, separators=(",", ": ")))
        else:
            filey.writelines(json.dumps(json_obj))
    return filename


def read_json(input_file):
    """Read json from an input file.

       Arguments:
         - input_file (str) : the filename to read
    """
    with open(input_file, "r") as filey:
        data = json.loads(filey.read())
    return data


def save_pickle(obj, filename):
    """Save a pickle to file

       Arguments:
         - obj (any) : the object to pickle
        - filename (str) : the output file to write to
    """
    with open(filename, "wb") as fd:
        pickle.dump(obj, fd)
    return filename


def mkdir_p(path):
    """mkdir_p attempts to get the same functionality as mkdir -p

       Arguments:
        - path (str) : the path to create
    """
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise e


def get_tmpfile(prefix=""):
    """get a temporary file with an optional prefix. By default, the file
       is closed (and just a name returned).

       Arguments:
        - prefix (str) : prefix with this string
    """
    tmpdir = tempfile.gettempdir()
    prefix = os.path.join(tmpdir, os.path.basename(prefix))
    fd, tmp_file = tempfile.mkstemp(prefix=prefix)
    os.close(fd)
    return tmp_file


def get_tmpdir(prefix="", create=True):
    """get a temporary directory for an operation.

       Arguments:
        - prefix (str) : prefix with this string
        - create (bool) : create the folder (defaults to true) 
    """
    tmpdir = tempfile.gettempdir()
    prefix = prefix or "rse-temp"
    prefix = "%s.%s" % (prefix, next(tempfile._get_candidate_names()))
    tmpdir = os.path.join(tmpdir, prefix)
    if not os.path.exists(tmpdir) and create:
        os.mkdir(tmpdir)
    return tmpdir



#!/usr/bin/env python
"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

import os
import sys
import pytest


@pytest.mark.parametrize("database", ["filesystem", "sqlite"])
def test_parsers_filesystem(tmp_path, database):
    """test each parser with the filesystem database.
    """
    from rse.main import Encyclopedia

    config_dir = os.path.join(str(tmp_path), "software")
    os.mkdir(config_dir)
    config_file = os.path.join(config_dir, "rse.ini")

    enc = Encyclopedia(config_file=config_file, generate=True, database="database")

    # Each uid should map to a parser
    uids = [["github", "github.com/singularityhub/sregistry"]]
    for i, parts in enumerate(uids):

        parser = parts[0]
        uid = parts[1]
        repo = enc.get_or_create(uid)

        assert repo.parser.name == parser
        assert repo.filename == os.path.join(
            enc.config_dir, "database", repo.uid, "metadata.json"
        )
        assert repo.summary()

        # repo.load includes the file dump, the upper level keys should be same
        content = repo.load()
        for key in ["parser", "uid", "url", "data"]:
            assert key in content

        # repo.export includes the executor specific data
        data = repo.export()
        assert "timestamp" in data


@pytest.mark.parametrize("database", ["filesystem", "sqlite"])
def test_filesystem(tmp_path, database):
    """Test loading and using a queue with the filesystem database.
    """
    from rse.main import Encyclopedia

    config_dir = os.path.join(str(tmp_path), "software")
    os.mkdir(config_dir)
    config_file = os.path.join(config_dir, "rse.ini")

    # System exit if doesn't exist
    with pytest.raises(SystemExit):
        enc = Encyclopedia(config_file=config_file)

    enc = Encyclopedia(config_file=config_file, generate=True, database=database)
    assert enc.config.configfile == config_file

    assert os.path.exists(config_dir)
    assert enc.config_dir == config_dir
    assert enc.config.configfile == os.path.join(enc.config_dir, "rse.ini")
    assert enc.database == database
    assert enc.db.database == database
    if database == "filesystem":
        assert enc.db.data_base == os.path.join(config_dir, "database")

    # Test list, empty without anything
    assert not enc.list()

    # Add a repo
    repo = enc.add("github.com/singularityhub/sregistry")
    assert len(enc.list()) == 1

    # enc.get should return last repo, given no id
    lastrepo = enc.get()
    assert lastrepo.uid == repo.uid

    # Summary
    enc.summary()
    enc.summary("github.com/singularityhub/sregistry")
    enc.analyze("github.com/singularityhub/sregistry")
    enc.analyze_bulk()

    # Clean up a specific repo (no prompt)
    enc.clear(repo.uid, noprompt=True)
    assert len(enc.list()) == 0
    enc.clear(noprompt=True)
    assert not enc.list()

    # Get the taxonomy or criteria
    enc.list_taxonomy()
    enc.list_criteria()



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from rse.utils.urls import get_user_agent
from rse.utils.file import mkdir_p, write_file, write_json
from rse.defaults import RSE_URL_PREFIX
import logging
import shutil
import os
import sys
import requests
import time

here = os.path.abspath(os.path.dirname(__file__))

bot = logging.getLogger("rse.app.export")


def export_web_static(export_dir, base_url, client, force=False):
    """Export a running web interface to a folder. If the folder exists, the
       user must use force. This should be run via:
           rse export --type static-web [export_dir]
       If the user manually starts the server, the user needs to do:
           export RSE_DISABLE_ANNOTATE=True before the server is started
       to disable the annotation interface button. This will be fixed
       in a future PR to have an interface that submits an issue to do
       an annotation, but this needs to be developed first.

       Arguments:
        - export_dir (str)      : the path to an export directory
        - base_url (str)        : the base url of the server, including port
        - client (Encyclopedia) : the encyclopedia to use
        - force (bool)          : if directory exists, overwrite
    """
    print(f"Starting export for {base_url}")
    time.sleep(2)

    # Ensure that the server is running
    try:
        requests.head(base_url).status_code == 200
    except:
        bot.info(
            "Please export after the server is running: export --type static-web [export_dir]"
        )
        return

    # Output directory cannot exist if force
    if os.path.exists(export_dir) and not force:
        sys.exit(f"{export_dir} exists, use --force to overwrite.")

    # Create export directory if it doesn't exist
    if not os.path.exists(export_dir):
        os.mkdir(export_dir)

    # Copy static files
    static_files = os.path.join(export_dir, "static")
    if not os.path.exists(static_files):
        shutil.copytree(os.path.join(here, "static"), static_files)

    # Prepare urls (and filepath relative to export_dir) for export
    urls = {base_url: "index.html"}

    # Create static data export
    data = []

    # Add repos and static annotation
    for repo in client.list():
        repo = client.get(repo[0])
        repo_path = os.path.join("repository", repo.uid)
        data.append(
            {
                "uid": repo.uid,
                "url": repo.url,
                "rel": "%s%s" % (RSE_URL_PREFIX, repo_path),
                "avatar": repo.avatar,
                "description": repo.description,
            }
        )
        urls["%s%s%s" % (base_url, RSE_URL_PREFIX, repo_path)] = os.path.join(
            repo_path, "index.html"
        )

        # Static annotation endpoints
        for annotation_type in ["criteria", "taxonomy"]:
            urls[
                "%s%s%s/annotate-%s"
                % (base_url, RSE_URL_PREFIX, repo_path, annotation_type)
            ] = os.path.join(repo_path, "annotate-%s" % annotation_type, "index.html")

    # Add API endpoints
    urls["%s%sapi/repos" % (base_url, RSE_URL_PREFIX)] = os.path.join(
        "api", "repos", "index.json"
    )
    urls["%s%sapi/taxonomy" % (base_url, RSE_URL_PREFIX)] = os.path.join(
        "api", "taxonomy", "index.json"
    )
    urls["%s%sapi/criteria" % (base_url, RSE_URL_PREFIX)] = os.path.join(
        "api", "criteria", "index.json"
    )

    for url, outfile in urls.items():

        # Skip if we've already created it
        if os.path.exists(outfile):
            continue

        # Update the output file with the repository
        outfile = os.path.join(export_dir, outfile)

        # Create nested output folder, if doesn't exist
        out_dir = os.path.dirname(outfile)
        if not os.path.exists(out_dir):
            mkdir_p(out_dir)

        # Url might have a prefix
        response = requests.get(url, headers={"User-Agent": get_user_agent()})
        if response.status_code == 200:
            write_file(outfile, response.text)
        else:
            print(f"Issue parsing {url}")

    print("Generating data export")
    write_json(data, os.path.join(export_dir, "data.json"))
    print("Export is complete!")



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from multiprocessing import Process
import os

from rse.main import Encyclopedia
import logging
from rse.utils.file import write_file

bot = logging.getLogger("rse.client")


def main(args, extra):

    client = Encyclopedia(config_file=args.config_file, database=args.database)

    # Case 1: empty list indicates listing all
    if os.path.exists(args.path) and not args.force:
        bot.error("{args.path} already exists, use --force to overwrite it.")

    # Export a list of repos
    if args.export_type == "repos-txt":
        # We just want the unique id, the first result
        repos = [x[0] for x in client.list()]
        write_file(args.path, "\n".join(repos))
        bot.info(f"Wrote {len(repos)} to {args.path}")

    # Static web export from flask to a directory
    elif args.export_type == "static-web":
        from rse.app.server import start
        from rse.app.export import export_web_static

        # Start the webserver on a separate process
        p = Process(
            target=start,  # port, debug, client, host, log-level, disable_annotate
            args=(args.port, args.debug, client, args.host, args.log_level, True),
        )
        p.start()

        # Do the export!
        export_web_static(
            export_dir=args.path,
            base_url="http://%s:%s" % (args.host, args.port),
            force=args.force,
            client=client,
        )

        # Ensure that it stops!
        p.kill()



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from rse.logger import RSE_LOG_LEVEL
from rse.exceptions import MissingEnvironmentVariable
import logging
import multiprocessing
import os


logging.basicConfig(level=getattr(logging, RSE_LOG_LEVEL))
bot = logging.getLogger("rse.defaults")


def getenv(variable_key, default=None, required=False, silent=True):
    """ attempt to get an environment variable. If the variable
        is not found, None is returned.

        Arguments:

         - variable_key (str) : the variable name
         - required (bool) : exit with error if not found
         - silent (bool) : Do not print debugging information
    """
    variable = os.environ.get(variable_key, default)
    if variable is None and required:
        raise MissingEnvironmentVariable(variable_key)

    if not silent and variable is not None:
        bot.debug("%s found as %s" % (variable_key, variable))

    return variable


RSE_NPROC = multiprocessing.cpu_count()
RSE_WORKERS = int(getenv("RSE_WORKERS", RSE_NPROC * 2 + 1))
RSE_SHELL = getenv("RSE_SHELL", "ipython")
RSE_CONFIG_FILE = getenv("RSE_CONFIG_FILE", "rse.ini")

# Default database is filesystem
RSE_DATABASE = getenv("RSE_DATABASE")

# Database folder for filesystem or sqlite database
RSE_DATABASE_STRING = os.environ.get("RSE_DATABASE")

# Parsers installed
RSE_PARSERS = ["github"]

# Default taxonomy and criteria endpoints, and place to post annotation issues
RSE_API_ENDPOINT = getenv("RSE_API_ENDPOINT", "https://rseng.github.io/rseng/api")
RSE_ISSUE_ENDPOINT = getenv("RSE_ISSUE_ENDPOINT", "https://github.com/rseng/software")

# MUST start and end with slash
RSE_URL_PREFIX = getenv("RSE_URL_PREFIX", "/")
if not RSE_URL_PREFIX.startswith("/"):
    RSE_URL_PREFIX = "/%s" % RSE_URL_PREFIX
if not RSE_URL_PREFIX.endswith("/"):
    RSE_URL_PREFIX = "%s/" % RSE_URL_PREFIX

# Dashboard settings
RSE_HOSTNAME = getenv("RSE_HOSTNAME", "127.0.0.1")



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from rse.utils.file import read_file

from datetime import datetime
import os
import json
import tempfile
import subprocess


class Capturing:
    """capture output from stdout and stderr into capture object.
       This is based off of github.com/vsoch/gridtest but modified
       to write files. The stderr and stdout are set to temporary files at
       the init of the capture, and then they are closed when we exit. This
       means expected usage looks like:

       with Capturing() as capture:
           process = subprocess.Popen(...)
           
       And then the output and error are retrieved from reading the files:
       and exposed as properties to the client:

           capture.out
           capture.err

       And cleanup means deleting these files, if they exist.
    """

    def __enter__(self):
        self.set_stdout()
        self.set_stderr()
        self.output = []
        self.error = []
        return self

    def set_stdout(self):
        self.stdout = open(tempfile.mkstemp()[1], "w")

    def set_stderr(self):
        self.stderr = open(tempfile.mkstemp()[1], "w")

    def __exit__(self, *args):
        self.stderr.close()
        self.stdout.close()

    @property
    def out(self):
        """Return output stream. Returns empty string if empty or doesn't exist.
           Returns (str) : output stream written to file
        """
        if os.path.exists(self.stdout.name):
            return read_file(self.stdout.name)
        return ""

    @property
    def err(self):
        """Return error stream. Returns empty string if empty or doesn't exist.
           Returns (str) : error stream written to file
        """
        if os.path.exists(self.stderr.name):
            return read_file(self.stderr.name)
        return ""

    def cleanup(self):
        for filename in [self.stdout.name, self.stderr.name]:
            if os.path.exists(filename):
                os.remove(filename)


class ParserBase:
    """A parser base exists to extract and format repository metadata.
    """

    name = "base"

    def __init__(self, uid=None):
        """set a unique id that includes parser name (type) and unique identifier)
        """
        self.uid = None
        if uid is not None:
            self.set_uid(uid)
        if not hasattr(self, "data"):
            self.data = {}

    def set_uid(self, uid):
        """Given a unique resource identifier, set it for the parser
        """
        uid = self._set_uid(uid)
        self.uid = "%s/%s" % (self.name, uid)

    def _set_uid(self, uid):
        """Given a uri from the user, parse the consistent identifier (e.g.,
           in the case of GitHub a repository username and name) 
        """
        raise NotImplementedError

    def load(self, data):
        """If a repository has already been instantiated, we might want to load
           data into a parser to interact with it
        """
        if isinstance(data, str):
            data = json.loads(data)
        self.data = data

    def _export_common(self):
        """export common repo variables such as timestamp when it was updated. 
           This might include envars at some point, but we'd need to be careful.
        """
        return {"timestamp": str(datetime.now())}

    def get_url(self, data):
        """a common function for a parser to return the html url for the
           upper level of metadata
        """
        raise NotImplementedError

    def get_description(self, data):
        """a common function for a parser to return a description.
        """
        raise NotImplementedError

    def export(self):
        """return data as json. This is intended to save to the software database.
           Any important parser specific metadata should be added to self.data
        """
        # Get common context (e.g., pwd)
        common = self._export_common()
        common.update(self.data)
        return common

    def get_metadata(self, uri, **kwargs):
        """The get_metadata function should take a general URI for a parser
           and populate the self.data
        """
        raise NotImplementedError

    def capture(self, cmd):
        """capture is a helper function to capture a shell command. We
           use Capturing and then save attributes like the pid, output, error
           to it, and return to the calling function. For example:

           capture = self.capture_command(cmd)
           self.pid = capture.pid
           self.returncode = capture.returncode
           self.out = capture.output
           self.err = capture.error
        """
        # Capturing provides temporary output and error files
        with Capturing() as capture:
            process = subprocess.Popen(
                cmd,
                stdout=capture.stdout,
                stderr=capture.stderr,
                universal_newlines=True,
            )
            capture.pid = process.pid
            returncode = process.poll()

            # Iterate through the output
            while returncode is None:
                returncode = process.poll()

        # Get the remainder of lines, add return code
        capture.output += [x for x in self.decode(capture.out) if x]
        capture.error += [x for x in self.decode(capture.err) if x]

        # Cleanup capture files and save final return code
        capture.cleanup()
        capture.returncode = returncode
        return capture

    def get_setting(self, key, default=None):
        """Get a setting, meaning that we first check the environment, then
           the config file, and then (if provided) a default.
        """
        # First preference to environment
        envar = ("RSE_%s_%s" % (self.name, key)).upper()
        envar = os.environ.get(envar)
        if envar is not None:
            return envar

        # Next preference to config setting
        parser = "parser.%s" % self.name

        # Parsers instantiated separate from database won't have config
        if not hasattr(self, "config"):
            return default
        if parser not in self.config.config:
            return default
        if key in self.config.config[parser]:
            return self.config.get(parser, key)
        return default

    def summary(self):
        if self.uid:
            return "[%s][%s]" % (self.name, self.uid)
        return "[%s]" % self.name



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

import flask
from flask_restful import Resource, Api
from rse.app.server import app
from rse.defaults import RSE_URL_PREFIX


def list_repos(parser=None):
    """A shared function to list one or more repos, optionally for a parser,
       and return a json list to serialize to the api view
    """
    repos = []
    url = flask.request.host_url
    for i, repo in enumerate(app.client.list(parser)):
        repos.append(
            {
                "uid": repo[0],
                "html_url": "%sparser/%s" % (url, repo[0]),
                "api_url": "%sapi/repos/%s" % (url, repo[0]),
            }
        )
    return repos


class apiList(Resource):
    """display all tasks
    """

    def get(self):
        return list_repos()


class apiListParser(Resource):
    """display all tasks for an executor
    """

    def get(self, parser):
        return list_repos(parser)


class apiGet(Resource):
    """display a specific repo
    :param uid: id for a specific task
    """

    def get(self, uid):
        repo = app.client.get(uid)
        return repo.export()


class apiEndpoints(Resource):
    """show all API endpoints
    """

    def get(self):
        url = flask.request.host_url
        return {
            "%sapi" % RSE_URL_PREFIX: "%sapi" % url,
            "%sapi/repos" % RSE_URL_PREFIX: "%sapi/repos" % url,
            "%sapi/criteria" % RSE_URL_PREFIX: "%sapi/criteria" % url,
            "%sapi/taxonomy" % RSE_URL_PREFIX: "%sapi/taxonomy" % url,
            "%sapi/repos/<path:uid>" % RSE_URL_PREFIX: "%sapi/repos/<path:uid>" % url,
            "%sapi/repos/parser/<string:parser>"
            % RSE_URL_PREFIX: "%sapi/repos/parser/<string:parser>"
            % url,
        }


class apiGetCriteria(Resource):
    """display the listing of criteria
    """

    def get(self):
        return app.client.list_criteria()


class apiGetTaxonomy(Resource):
    """display the flattened taxonomy.
    """

    def get(self):
        return app.client.list_taxonomy()


api = Api(app)
api.add_resource(apiEndpoints, "%sapi" % RSE_URL_PREFIX)
api.add_resource(apiList, "%sapi/repos" % RSE_URL_PREFIX)
api.add_resource(apiGet, "%sapi/repos/<path:uid>" % RSE_URL_PREFIX)
api.add_resource(apiGetTaxonomy, "%sapi/taxonomy" % RSE_URL_PREFIX)
api.add_resource(apiGetCriteria, "%sapi/criteria" % RSE_URL_PREFIX)
api.add_resource(apiListParser, "%sapi/repos/parser/<string:parser>" % RSE_URL_PREFIX)



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

import json


try:
    from sqlalchemy import Column, DateTime, String, Text, func
    from sqlalchemy.ext.declarative import declarative_base
except:
    raise ImportError("sqlalchemy", "is required for a non-filesystem database.")

# Shared declarative base
Base = declarative_base()


class SoftwareRepository(Base):
    """A software repository.
    """

    __tablename__ = "software_repository"
    uid = Column(String(150), primary_key=True)
    timestamp = Column(DateTime, default=func.now())
    parser_name = Column(String(50))
    data = Column(Text, nullable=True)
    taxonomy = Column(Text, nullable=True)
    criteria = Column(Text, nullable=True)

    def __init__(self, uid=None, parser=None, data={}):
        self.uid = uid
        self.parser_name = parser
        self.data = data

    def summary(self):
        return self.parser.summary()

    @property
    def url(self):
        data = json.loads(self.data)
        return self.parser.get_url(data)

    @property
    def avatar(self):
        data = json.loads(self.data)
        return self.parser.get_avatar(data)

    @property
    def description(self):
        data = json.loads(self.data)
        return self.parser.get_description(data)

    def load(self):
        """loading a software repository means exporting as json"""
        data = {"parser": self.parser_name, "uid": self.uid, "data": {}}

        if self.data:
            data["data"] = json.loads(self.data)

        return data

    def get_criteria(self):
        """load criteria into a dictionary"""
        return json.loads(self.criteria or "{}")

    def get_taxonomy(self):
        """load taxonomy into a dictionary"""
        return json.loads(self.taxonomy or "{}")

    def update_criteria(self, uid, username, response):
        """Given a username and unique id update criteria"""
        criteria = self.get_criteria()

        if uid not in criteria:
            criteria[uid] = {}
        if response:
            criteria[uid][username] = response
            self.criteria = json.dumps(criteria)

    def update_taxonomy(self, username, uids):
        """Given a username and unique id update taxonomy items"""
        taxonomy = self.get_taxonomy()

        if username not in taxonomy:
            taxonomy[username] = uids
        self.taxonomy = json.dumps(taxonomy)

    def export(self):
        """Export removes the outer wrapper, and just returns the data"""
        return self.load().get("data", {})

    def __repr__(self):
        return "<SoftwareRepository %r>" % self.uid

    # Annotation

    def has_criteria_annotation(self, uid, username):
        """Determine if a repository has been annotated by a user.
        """
        if not self.criteria:
            return False
        criteria = json.loads(self.criteria)
        if uid not in criteria:
            return False
        if username not in criteria[uid].get("users", []):
            return False
        return True

    def has_taxonomy_annotation(self, username):
        """Determine if a repository has been annotated by a user.
        """
        if not self.taxonomy:
            return False
        taxonomy = json.loads(self.taxonomy)
        if username not in taxonomy:
            return False
        return True



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from rse.utils.urls import get_user_agent, check_response, repository_regex
from rse.main.parsers import get_parser
import logging
import requests
import re

from .base import ScraperBase

bot = logging.getLogger("rse.main.scrapers.hal")

# Allow the regex to have newline at end
repository_regex = repository_regex.strip("$")


class HalScraper(ScraperBase):

    name = "hal"

    def __init__(self, query=None, **kwargs):
        super().__init__(query)

    def latest(self, paginate=False):
        """populate self.results with some number of latest entries. Unlike 
           a search, a latest scraper does not by default paginate. Hal will by
           default return all entries, so the user is required to define a number
           for latest.
        """
        url = "https://api.archives-ouvertes.fr/search/?q=github&fq=docType_s:(SOFTWARE)&wt=json"
        return self.scrape(url)

    def search(self, query, paginate=True):
        """populate self.results with a listing based on matching a search criteria.
           we search the description.
        """
        url = (
            "http://api.archives-ouvertes.fr/search/?q=%s&fq=docType_s:(SOFTWARE)&wt=json"
            % query
        )
        return self.scrape(url)

    def scrape(self, url, paginate=False):
        """A shared function to scrape a set of repositories. Since the JoSS
           pages for a search and the base are the same, we can use a shared
           function.
        """
        # Api doesn't appear to have pagination
        response = requests.get(url, headers={"User-Agent": get_user_agent()})
        data = check_response(response)

        for entry in data.get("response", {}).get("docs", []):
            page_url = entry["uri_s"]
            response = requests.get(page_url, headers={"User-Agent": get_user_agent()})
            repo_url = None
            if response.status_code == 200:
                match = re.search(repository_regex, response.text, re.IGNORECASE)
                if match:
                    repo_url = match.group()

            if repo_url:
                bot.info("Found repository: %s" % repo_url)
                self.results.append(repo_url)

        return self.results

    def create(self, database=None, config_file=None):
        """After a scrape (whether we obtain latest or a search query) we
           run create to create software repositories based on results.
        """
        from rse.main import Encyclopedia

        client = Encyclopedia(config_file=config_file, database=database)
        for repo_id in self.results:
            repo = get_parser(repo_id)

            # Add results that don't exist
            if not client.exists(repo.uid):
                client.add(repo.uid)



#!/usr/bin/env python

"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from rse.logger import RSE_LOG_LEVEL, RSE_LOG_LEVELS
from rse.defaults import RSE_CONFIG_FILE
import rse
import argparse
import sys
import logging


def get_parser():
    parser = argparse.ArgumentParser(
        description="Research software engineering software inspector."
    )

    parser.add_argument(
        "--version",
        dest="version",
        help="suppress additional output.",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--log_level",
        dest="log_level",
        choices=RSE_LOG_LEVELS,
        default=RSE_LOG_LEVEL,
        help="Customize logging level for rse inspector.",
    )

    # Configuration file
    parser.add_argument(
        "--config_file",
        dest="config_file",
        default=RSE_CONFIG_FILE,
        help="Path to rse.ini configuration file.",
    )

    description = "actions for rse"
    subparsers = parser.add_subparsers(
        help="rse actions", title="actions", description=description, dest="command",
    )

    # print version and exit
    subparsers.add_parser("version", help="show software version")

    # Annotate criteria or taxonomy
    annotate = subparsers.add_parser(
        "annotate", help="Annotate a database with criteria or taxonomy"
    )
    annotate.add_argument(
        "type",
        help="Type to annotate (taxonomy or criteria)",
        nargs=1,
        choices=["taxonomy", "criteria"],
    )
    annotate.add_argument(
        "-u",
        "--username",
        dest="username",
        default=None,
        help="GitHub username (must be provided if not available with git config)",
    )
    annotate.add_argument(
        "-r",
        "--repo",
        dest="repo",
        default=None,
        help="Specify a particular repository name to annotate.",
    )
    annotate.add_argument(
        "-f",
        "--file",
        dest="file",
        default=None,
        help="Read annotation from file (e.g., markdown list, GitHub issue template).",
    )
    annotate.add_argument(
        "--all",
        "-a",
        dest="all_repos",
        help="Annotate all repos, even those that have already been seen (defaults to show only those unseen).",
        default=False,
        action="store_true",
    )

    # Generate a key for the interface
    generate = subparsers.add_parser(
        "generate-key",
        help="generate a key for rse start, should be exported to RSE_SERVER_KEY.",
    )

    # Init
    init = subparsers.add_parser(
        "init", help="Add an rse.ini to the present working directory."
    )
    init.add_argument(
        "path", help="Path to generate rse.ini file", nargs="?", default=".",
    )

    # Config
    config = subparsers.add_parser(
        "config", help="Update an rse.ini configuration file."
    )

    # Clear
    clear = subparsers.add_parser("clear", help="Remove software from the database.")
    clear.add_argument("target", nargs="?")
    clear.add_argument(
        "--force",
        dest="force",
        help="Don't ask for confirmation for delete (for headless).",
        default=False,
        action="store_true",
    )

    # Exists
    exists = subparsers.add_parser(
        "exists", help="Determine if an entry exists in the database."
    )

    # Export
    export = subparsers.add_parser(
        "export", help="Export repository names, metadata, or static files."
    )
    export.add_argument(
        "--type",
        dest="export_type",
        help="Type to export (defaults to repos.txt list)",
        default="repos-txt",
        choices=["repos-txt", "static-web"],
    )
    export.add_argument(
        "--force",
        dest="force",
        help="Don't ask for confirmation to overwrite existing file(s).",
        default=False,
        action="store_true",
    )
    export.add_argument(
        "path",
        help="Fileame to export repos to (default repos.txt)",
        default="repos.txt",
    )

    # Metrics
    summary = subparsers.add_parser(
        "summary", help="View summary metrics for all repositories."
    )
    summary.add_argument(
        "--type",
        dest="metric_type",
        help="Metric type to view.",
        default="summary",
        choices=["criteria", "taxonomy", "users"],
    )
    summary.add_argument(
        "repo", help="Filter down to one repository.", default=None, nargs="?"
    )

    analyze = subparsers.add_parser(
        "analyze", help="View metrics for a specific repository."
    )
    analyze.add_argument(
        "repo", help="Software repository to show", default=None,
    )
    analyze.add_argument(
        "--ct",
        "--cthresh",
        dest="cthresh",
        help="Criteria threshold (between 0 and 1)",
        default=0.5,
        type=percentage_type_asint,
    )
    analyze.add_argument(
        "--tt",
        "--tthresh",
        dest="tthresh",
        help="Minimum taxonomy votes to count category in response",
        default=1,
        type=positive_int_type,
    )

    # Update
    update = subparsers.add_parser(
        "update", help="Update one or more software entries."
    )

    update.add_argument(
        "-p",
        "--path",
        dest="path",
        default=None,
        help="Path to single folder or set of folders to update.",
    )
    update.add_argument(
        "--force",
        dest="force",
        help="If a repository is not present, add it.",
        default=False,
        action="store_true",
    )

    # List repos and print to terminal
    ls = subparsers.add_parser("ls", help="List software")
    ls.add_argument(
        "parser", help="list one or more parsers or specific software.", nargs="*"
    )

    # Search for software
    search = subparsers.add_parser(
        "search", help="Search for a piece of research software",
    )
    search.add_argument("query", nargs="*")

    # Scrape for new repos
    scrape = subparsers.add_parser(
        "scrape", help="Add new software repositories from a resource.",
    )
    scrape.add_argument("scraper_name", nargs=1)
    scrape.add_argument("query", nargs="?")
    scrape.add_argument(
        "--dry-run",
        dest="dry_run",
        help="Dump scrape output to the terminal, but don't create new repos.",
        default=False,
        action="store_true",
    )

    # Shell
    subparsers.add_parser(
        "shell", help="start an interactive shell for an encyclopedia"
    )

    # Start the rse dashboard
    start = subparsers.add_parser(
        "start", help="start an interface to browse software (requires Flask)"
    )
    for command in [start, export]:
        command.add_argument(
            "--port",
            dest="port",
            default=5000,
            type=int,
            help="select port to run dashboard on (defaults to 5000)",
        )
        command.add_argument(
            "--host",
            dest="host",
            default="127.0.0.1",
            type=str,
            help="the hostname to run for the server (defaults to 127.0.0.1)",
        )
        command.add_argument(
            "--debug",
            dest="debug",
            help="run server in debug mode (defaults to False)",
            default=False,
            action="store_true",
        )
        command.add_argument(
            "--disable-annotate",
            dest="disable_annotate",
            help="disable annotation button (defaults to False)",
            default=False,
            action="store_true",
        )

    # Label a repository with metadata, e.g., add a DOI.
    label = subparsers.add_parser(
        "label", help="Add a metadata value for an existing repository"
    )
    label.add_argument("values", nargs=3)
    label.add_argument(
        "--force",
        dest="force",
        help="Overwrite existing label, if it exists.",
        default=False,
        action="store_true",
    )

    # Print complete metadata for a specific piece of software
    get = subparsers.add_parser("get", help="Show metadata for software")
    add = subparsers.add_parser("add", help="Add a repository to the database.")

    # Most commands allow specification of a database backend
    for command in [
        add,
        analyze,
        annotate,
        clear,
        config,
        exists,
        export,
        get,
        init,
        label,
        ls,
        search,
        scrape,
        start,
        summary,
        update,
    ]:
        command.add_argument(
            "--database",
            dest="database",
            choices=["filesystem", "sqlite"],
            default=None,
            help="database backend to use, to override configuration.",
        )

    for command in [exists, get]:
        command.add_argument("uid", help="uri within software namespace.", nargs="?")

    for command in [update, add]:
        command.add_argument("uid", help="uri within software namespace.", nargs="?")
        command.add_argument(
            "--file",
            dest="file",
            default=None,
            help="single line delimited file of repositories.",
        )

    return parser


def percentage_type_asint(arg):
    """ensure that an input is between 0 and 1"""
    try:
        number = float(arg)
    except ValueError:
        raise argparse.ArgumentTypeError("Must be a floating point number")
    if number < 0 or number > 1:
        raise argparse.ArgumentTypeError("Argument must be between 0 and 1")
    return number


def positive_int_type(arg):
    """ensure user is providing a positive integer"""
    value = int(arg)
    if value < 0:
        raise argparse.ArgumentTypeError(
            "%s is an invalid positive integer value" % arg
        )
    return value


def main():
    """main entrypoint for rse
    """

    parser = get_parser()

    def help(return_code=0):
        """print help, including the software version and active client 
           and exit with return code.
        """
        version = rse.__version__

        print("\nResearch Software Engineering software inspector v%s" % version)
        parser.print_help()
        sys.exit(return_code)

    # If the user didn't provide any arguments, show the full help
    if len(sys.argv) == 1:
        help()

    # If an error occurs while parsing the arguments, the interpreter will exit with value 2
    args, extra = parser.parse_known_args()

    # Set the logging level
    logging.basicConfig(level=getattr(logging, args.log_level))
    bot = logging.getLogger("rse.client")
    bot.setLevel(getattr(logging, args.log_level))

    # Show the version and exit
    if args.command == "version" or args.version:
        print(rse.__version__)
        sys.exit(0)

    # Does the user want a shell?
    if args.command == "analyze":
        from .metrics import analyze as main
    if args.command == "annotate":
        from .annotate import main
    if args.command == "add":
        from .add import main
    if args.command == "clear":
        from .clear import main
    if args.command == "config":
        from .config import main
    if args.command == "exists":
        from .exists import main
    if args.command == "export":
        from .export import main
    if args.command == "generate-key":
        from .generate import main
    if args.command == "label":
        from .label import main
    if args.command == "update":
        from .update import main
    if args.command == "get":
        from .get import main
    if args.command == "init":
        from .init import main
    if args.command == "ls":
        from .listing import main
    if args.command == "summary":
        from .metrics import summary as main
    if args.command == "search":
        from .search import main
    if args.command == "scrape":
        from .scrape import main
    if args.command == "shell":
        from .shell import main
    if args.command == "start":
        from .start import main

    # Pass on to the correct parser
    main(args=args, extra=extra)


if __name__ == "__main__":
    main()



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

import random
import logging

bot = logging.getLogger("rse.utils.urls")

repository_regex = (
    "(github.com|gitlab.com)/(?P<owner>[\w,\-,\_]+)/(?P<repo>[\w,\-,\_]+)$"
)


def check_response(response):
    """Given a requests.Response, return the json data if it's successful.
       Otherwise issue an error and return None
    """
    if response.status_code == 200:
        return response.json()

    elif response.status_code == 404:
        bot.error(f"Cannot find endpoint {response.url}.")

    elif response.status_code in [400, 401, 403]:
        bot.error(f"Permission denied to query {response.url}")

    else:
        bot.error(
            f"Cannot get {response.url}: {response.status_code}, {response.reason}"
        )
    return None


def get_user_agent():
    """
    Return a randomly chosen user agent for requests

    Returns:
        user agent string to include with User-Agent.
    """
    agents = [
        (
            "Mozilla/5.0 (X11; Linux x86_64) "
            "AppleWebKit/537.36 (KHTML, like Gecko) "
            "Chrome/57.0.2987.110 "
            "Safari/537.36"
        ),  # chrome
        (
            "Mozilla/5.0 (X11; Linux x86_64) "
            "AppleWebKit/537.36 (KHTML, like Gecko) "
            "Chrome/61.0.3163.79 "
            "Safari/537.36"
        ),  # chrome
        (
            "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:55.0) "
            "Gecko/20100101 "
            "Firefox/55.0"
        ),  # firefox
        (
            "Mozilla/5.0 (X11; Linux x86_64) "
            "AppleWebKit/537.36 (KHTML, like Gecko) "
            "Chrome/61.0.3163.91 "
            "Safari/537.36"
        ),  # chrome
        (
            "Mozilla/5.0 (X11; Linux x86_64) "
            "AppleWebKit/537.36 (KHTML, like Gecko) "
            "Chrome/62.0.3202.89 "
            "Safari/537.36"
        ),  # chrome
        (
            "Mozilla/5.0 (X11; Linux x86_64) "
            "AppleWebKit/537.36 (KHTML, like Gecko) "
            "Chrome/63.0.3239.108 "
            "Safari/537.36"
        ),  # chrome
    ]
    return random.choice(agents)



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from rse.utils.urls import get_user_agent, check_response, repository_regex
from rse.main.parsers import get_parser
import logging
import requests
import random
import re
from time import sleep

from .base import ScraperBase

bot = logging.getLogger("rse.main.scrapers.biotools")


class BioToolsScraper(ScraperBase):

    name = "biotools"

    def __init__(self, query=None, **kwargs):
        super().__init__(query)

    def latest(self, paginate=False):
        """populate self.results with some number of latest entries. Unlike 
           a search, a latest scraper does not by default paginate. The user 
           needs to interact directly with the Python client to scrape all.
        """
        url = "https://bio.tools/api/tool/?format=json"
        return self.scrape(url, paginate=paginate)

    def search(self, query, paginate=True):
        """populate self.results with a listing based on matching a search criteria.
           we search the description.
        """
        url = 'https://bio.tools/api/t/?description="%s"&format=json' % query
        return self.scrape(url, paginate=paginate)

    def scrape(self, url, paginate=False):
        """A shared function to scrape a set of repositories. Since the JoSS
           pages for a search and the base are the same, we can use a shared
           function.
        """
        # Handle pagination
        original_url = url
        while url is not None:

            response = requests.get(url, headers={"User-Agent": get_user_agent()})
            data = check_response(response)

            # Reset the url to be None
            url = None
            if data.get("next") and paginate:
                url = original_url + "&page=%s" % data.get("next", "").replace(
                    "?page=", ""
                )

            for entry in data.get("list", []):

                # Look for GitHub / GitLab URL
                repo = {}
                for link in entry.get("link", []):
                    if "Repository" in link["type"] and re.search(
                        repository_regex, link["url"], re.IGNORECASE
                    ):
                        repo["url"] = link["url"]

                # If we don't have a repository, search the homepage
                if not repo.get("url") and re.search(
                    repository_regex, entry["homepage"]
                ):
                    repo["url"] = entry["homepage"]

                # We must have a repository url to parse
                if not repo.get("url"):
                    continue

                # Look for a doi
                for pub in entry["publication"]:
                    if pub.get("doi"):
                        repo["doi"] = pub.get("doi")

                bot.info("Found repository: %s" % repo["url"])
                self.results.append(repo)

                # Sleep for a random amount of time to give a rest!
                sleep(random.choice(range(1, 10)) * 0.1)

        return self.results

    def create(self, database=None, config_file=None):
        """After a scrape (whether we obtain latest or a search query) we
           run create to create software repositories based on results.
        """
        from rse.main import Encyclopedia

        client = Encyclopedia(config_file=config_file, database=database)
        for result in self.results:
            uid = result["url"].split("//")[-1]
            repo = get_parser(uid)

            # Add results that don't exist
            if not client.exists(repo.uid):
                client.add(repo.uid)
                if result.get("doi") is not None:
                    client.label(repo.uid, key="doi", value=result.get("doi"))



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from rse.main import Encyclopedia
from rse.defaults import RSE_SHELL


def main(args, extra):

    lookup = {"ipython": ipython, "python": python, "bpython": bpython}
    shells = ["ipython", "python", "bpython"]

    # Provide shell if available
    shell = RSE_SHELL.lower()
    if shell in lookup:
        return lookup[shell](args)

    # Otherwise present order of liklihood to have on system
    for shell in shells:
        return lookup[shell](args)


def ipython(args):
    """give the user an ipython shell, optionally with an endpoint of choice.
    """
    client = Encyclopedia(config_file=args.config_file)
    assert client
    from IPython import embed

    embed()


def bpython(args):
    import bpython

    client = Encyclopedia(config_file=args.config_file)
    assert client
    bpython.embed(locals_={"client": client})


def python(args):
    import code

    client = Encyclopedia(config_file=args.config_file)
    assert client

    code.interact(local={"client": client})



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from rse.utils.urls import get_user_agent, check_response
from rse.main.parsers import get_parser
import logging
import requests

from .base import ScraperBase

bot = logging.getLogger("rse.main.scrapers.rsnl")

# Research Software Netherlands
class RSNLScraper(ScraperBase):

    name = "rsnl"

    def __init__(self, query=None, **kwargs):
        super().__init__(query)

    def latest(self, paginate=False):
        """The scraper should expose a function to populate self.results with
           some number of latest entries. Unlike a search, a latest scraper does
           not by default paginate.
        """
        url = "https://research-software.nl/api/software"
        return self.scrape(url)

    def search(self, query, paginate=True):
        """The scraper should expose a function to populate self.results with
           a listing based on matching a search criteria.
        """
        # Haven't figured out a way to search
        bot.warning(
            "The RSNL scraper does not support search, returning latest instead."
        )
        return self.latest(paginate)

    def scrape(self, url, paginate=False):
        """A shared function to scrape a set of repositories. Since the JoSS
           pages for a search and the base are the same, we can use a shared
           function.
        """
        response = requests.get(url, headers={"User-Agent": get_user_agent()})
        data = check_response(response) or []

        for entry in data:

            # I only see GitHub urls
            repo_url = entry.get("repositoryURLs", {}).get("github")
            repo_url = repo_url[0] if repo_url else None
            doi = entry.get("conceptDOI")
            doi = doi if doi and "FIXME" not in doi else None
            if repo_url and doi:
                bot.info("Found repository: %s" % repo_url)
                self.results.append({"url": repo_url, "doi": doi})
            elif repo_url:
                bot.info("Found repository: %s" % repo_url)
                self.results.append({"url": repo_url})

        return self.results

    def create(self, database=None, config_file=None):
        """After a scrape (whether we obtain latest or a search query) we
           run create to create software repositories based on results.
        """
        from rse.main import Encyclopedia

        client = Encyclopedia(config_file=config_file, database=database)
        for result in self.results:
            uid = result["url"].split("//")[-1]
            repo = get_parser(uid)

            # Add results that don't exist
            if not client.exists(repo.uid):
                client.add(repo.uid)
                if result.get("doi"):
                    client.label(repo.uid, key="doi", value=result.get("doi"))



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

import logging
import requests
import re

from .base import ParserBase

bot = logging.getLogger("rse.main.parsers.zenodo")


class ZenodoParser(ParserBase):

    name = "zenodo"
    matchstring = "^10[.][0-9]{4}/zenodo[.][0-9]{7}$"

    def __init__(self, uid=None, **kwargs):
        super().__init__(uid)

    def _set_uid(self, uid):
        """Given some kind of Zenodo uri, parse it
        """
        match = re.search(self.matchstring, uid)
        if not match:
            raise RuntimeError(f"{uid} does match a Zenodo DOI.")
        return match.group()

    def load_secrets(self):
        """load secrets, namely the access token, check if required and
           exit if not provided
        """
        self.token = self.get_setting("TOKEN")

    def get_url(self, data=None):
        """a common function for a parser to return the html url for the
           upper level of metadata
        """
        data = data or self.data
        return data.get("links", {}).get("html")

    def get_avatar(self, data=None):
        """a common function for a parser to return an image.
        """
        data = data or self.data
        return data.get("links", {}).get("badge")

    def get_description(self, data=None):
        """a common function for a parser to return a description.
        """
        data = data or self.data
        return data.get("metadata", {}).get("description")

    def get_metadata(self, uri=None, require_repo=True):
        """Retrieve repository metadata. The common metadata (timestamp) is
           added by the software repository parser, and here we need to
           ensure that the url field is populated with a correct url.

           Arguments:
           uri (str) : a repository uri string to override one currently set
           require_repo (bool) : require a repository to parse.
        """
        from rse.main.parsers import get_parser
        from rse.utils.urls import repository_regex

        repository_regex = repository_regex.rstrip("$")

        if uri:
            self.set_uri(uri)
        self.load_secrets()

        # Get the record number from the doi
        record = self.uid.split("/")[-1].replace("zenodo.", "")

        # Token isn't required for public entries
        if self.token:
            response = requests.get(
                "https://zenodo.org/api/records/%s" % record,
                json={"access_token": self.token},
            )
        else:
            response = requests.get("https://zenodo.org/api/records/%s" % record)

        # Successful query!
        if response.status_code == 200:
            self.data = response.json()

            # For Zenodo, we require a GitHub or GitLab related identifier to add
            repo_url = None
            for identifier in self.data["metadata"].get("related_identifiers", []):
                match = re.search(repository_regex, identifier["identifier"])
                if match:
                    repo_url = "https://%s" % match.group()
                    break

            # If we return None, the entry is not added
            if repo_url is None and require_repo is True:
                bot.warning(
                    "Repository url not found with Zenodo record, skipping add."
                )
                return repo_url

            # Convert the class into another parser type
            elif repo_url is not None:
                uid = self.uid
                self = get_parser(repo_url)
                self.get_metadata()
                self.data["doi"] = uid
                return self
            return self.data

        elif response.status_code == 404:
            bot.error(f"Cannot find doi {self.uid}.")

        elif response.status_code in [400, 401, 403]:
            bot.error(f"Permission denied to query {self.uid}")

        else:
            bot.error(
                f"Cannot get doi {self.uid}: {response.status_code}, {response.reason}"
            )
        return None



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""


class MissingEnvironmentVariable(RuntimeError):
    """Thrown if a required environment variable is not provided.
    """

    def __init__(self, varname, *args, **kwargs):
        super(MissingEnvironmentVariable, self).__init__(*args, **kwargs)
        self.varname = varname

    def __str__(self):
        return "Missing environment variable '{}' is required".format(self.varname)


class DirectoryNotFoundError(FileNotFoundError):
    """Thrown if a directory is not found
    """

    def __init__(self, dirname, reason, *args, **kwargs):
        super(DirectoryNotFoundError, self).__init__(*args, **kwargs)
        self.dirname = dirname
        self.reason = reason

    def __str__(self):
        return "{} {}.".format(self.dirname, self.reason)


class MissingDatabaseString(RuntimeError):
    """Thrown if a database string is required and not provided
    """

    def __init__(self, reason=None, *args, **kwargs):
        super(MissingDatabaseString, self).__init__(*args, **kwargs)
        self.reason = reason

    def __str__(self):
        return (
            self.reason
            or "A database url must be defined to use a relational database. Set with rse config --database"
        )


class DatabaseStringFormatError(RuntimeError):
    """Thrown if database prefix is not supported
    """

    def __str__(self):
        return (
            "Database must start with sqlite, filesystem, mysql+pymysql, or postgres."
        )


## Repos


class RepoError(RuntimeError):
    """Abstract base class for any kind of RepoError.
    """

    def __init__(self, uid=None, reason=None, *args, **kwargs):
        super(RepoError, self).__init__(*args, **kwargs)
        self.uid = uid or ""
        self.reason = reason or "There was a problem with repository"

    def __str__(self):
        return "{} {}".format(self.reason, self.uid)


class MultipleReposExistError(RepoError):
    """Thrown if multiple repos exist.
    """

    def __init__(self, uid, *args, **kwargs):
        reason = "More than one repository found for"
        super(MultipleReposExistError, self).__init__(
            uid=uid, reason=reason, *args, **kwargs
        )


class RepoNotFoundError(RepoError):
    """Thrown if a repo does not exist (for a remote)
    """

    def __init__(self, uid, *args, **kwargs):
        reason = "Cannot find repo"
        super(RepoNotFoundError, self).__init__(uid=uid, reason=reason, *args, **kwargs)


class NoReposError(RepoError):
    """Thrown if repos are requested, but there are none
    """

    def __init__(self, *args, **kwargs):
        reason = "There are no repos in the database."
        super(NoReposError, self).__init__(reason=reason, *args, **kwargs)


class RepoMetadataExistError(RepoError):
    """Thrown if a metadata value (label) already exists.
    """

    def __init__(self, uid, key, *args, **kwargs):
        reason = "Metadata value %s already is defined." % key
        super(RepoMetadataExistError, self).__init__(
            uid=uid, reason=reason, *args, **kwargs
        )



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from flask import render_template, request, redirect
from rse.app.server import app
from rse.defaults import RSE_URL_PREFIX, RSE_ISSUE_ENDPOINT
import random

## Repository Views


@app.route("%srepository/<path:uid>" % RSE_URL_PREFIX)
def repository_view(uid):

    # Obtain the repository and load the data.
    repo = app.client.get(uid)
    repo.parser.load(repo.data)

    if repo.parser.name == "github":
        skips = ["owner", "organization", "node_id"]
        return render_template(
            "parsers/github.html",
            repo=repo.load(),
            database=app.client.database,
            url_prefix=RSE_URL_PREFIX,
            skips=skips,
        )
    elif repo.parser.name == "gitlab":
        avatar = repo.parser.get_avatar()
        return render_template(
            "parsers/gitlab.html",
            repo=repo.load(),
            url_prefix=RSE_URL_PREFIX,
            database=app.client.database,
            avatar=avatar,
        )


@app.route("%sannotate" % RSE_URL_PREFIX)
def annotate_repos(message=""):
    return render_template(
        "annotate/repos.html",
        database=app.client.database,
        message=message,
        url_prefix=RSE_URL_PREFIX,
    )


@app.route("%sannotate-criteria" % RSE_URL_PREFIX, methods=["POST", "GET"])
def annotate_criteria():

    # If it's a post, update the annotation
    username = None
    if request.method == "POST":
        username = update_criteria()

    # Username on GitHub is required for annotation lookup
    username = request.args.get("username", username)
    print("Username is %s" % username)
    if username in [None, ""]:
        return annotate_repos(
            message="Please provide your GitHub username above to start the annotation session."
        )

    # Get the first set of criteria
    last = None
    annotation_sets = []
    for repo, criteria in app.client.yield_criteria_annotation_repos(
        username=username, unseen_only=True
    ):
        if last is not None and last.uid != repo.uid:
            break
        last = repo
        annotation_sets.append((repo, criteria,))

    return render_template(
        "annotate/criteria.html",
        database=app.client.database,
        sets=annotation_sets,
        username=username,
        url_prefix=RSE_URL_PREFIX,
    )


@app.route("%srepository/<path:uid>/annotate-criteria" % RSE_URL_PREFIX)
def annotate_static_criteria(uid):

    # Get criteria / annotation set for specific repository
    username = request.args.get("username")
    repo = app.client.get(uid)
    criteria = app.client.list_criteria()

    return render_template(
        "annotate/criteria-static.html",
        database=app.client.database,
        sets=criteria,
        repo=repo,
        username=username,
        issue_endpoint=RSE_ISSUE_ENDPOINT,
        url_prefix=RSE_URL_PREFIX,
    )


@app.route("%srepository/<path:uid>/annotate-taxonomy" % RSE_URL_PREFIX)
def annotate_static_taxonomy(uid):

    # Get criteria / annotation set for specific repository
    username = request.args.get("username")
    repo = app.client.get(uid)

    # If we don't have a color lookup, make one
    if not hasattr(app, "taxonomy"):
        app.taxonomy = generate_taxonomy(app)

    return render_template(
        "annotate/taxonomy-static.html",
        database=app.client.database,
        sets=app.taxonomy,
        repo=repo,
        username=username,
        issue_endpoint=RSE_ISSUE_ENDPOINT,
        url_prefix=RSE_URL_PREFIX,
    )


@app.route("%sannotate-taxonomy" % RSE_URL_PREFIX, methods=["GET", "POST"])
def annotate_taxonomy():

    # If we don't have a color lookup, make one
    if not hasattr(app, "taxonomy"):
        app.taxonomy = generate_taxonomy(app)

    # If it's a post, update the annotation
    username = None
    if request.method == "POST":
        username = update_taxonomy()

    # Username on GitHub is required for annotation lookup
    username = request.args.get("username", username)
    print("Username is %s" % username)
    if username in [None, ""]:
        return annotate_repos(
            message="Please provide your GitHub username above to start the annotation session."
        )

    # Generator for repos that need annotation based on username
    generator = app.client.yield_taxonomy_annotation_repos(username, unseen_only=True)

    # Return the taxonomy with one repo, until we run out
    try:
        repo = next(generator)
        print(repo)
        return render_template(
            "annotate/taxonomy.html",
            database=app.client.database,
            taxonomy=app.taxonomy,
            repo=repo,
            username=username,
            url_prefix=RSE_URL_PREFIX,
        )

    except StopIteration as exc:
        return annotate_repos(
            message="You have already annotated taxonomy items for all the repos!"
        )


# Helper Functions


def generate_taxonomy(app):

    taxonomy = app.client.list_taxonomy()

    # Update the color list with existing colors (consistency)
    colors = {}
    for item in taxonomy:
        colors[item["name"]] = item["color"]

    # Add parents to the lookup
    colors["Software to directly conduct research"] = "darkblue"
    colors["General software"] = "darkgoldenrod"
    colors["Software to support research"] = "darkgreen"
    colors["Incidentally used for research"] = "darkmagenta"
    colors["Used for research but not explicitly for it"] = "darkslateblue"
    colors["Domain-specific software"] = "darkcyan"
    colors["Explicitly for research"] = "mediumpurple"

    # Generate a color for each unique entry
    for entry in taxonomy:
        parts = [x.strip() for x in entry["path"].split(">>")]
        colorlist = []
        for part in parts:
            colorlist.append(colors[part])
        entry["colors"] = colorlist
    return taxonomy


def update_criteria():

    updates = {}
    repo_uid = request.form.get("repo_uid")
    username = request.form.get("username")
    for key in request.form.keys():
        for value in request.form.getlist(key):
            if key.startswith("radio-RSE-"):
                updates[key.replace("radio-", "", 1)] = value

    # Do the update for each criteria
    repo = app.client.get(repo_uid)
    for key, response in updates.items():
        print(f"Updating {repo} {key} with {response}")
        repo.update_criteria(key, username, response)

    # The filesystem database saves at the end
    app.client.save_criteria(repo)

    return username


def update_taxonomy():

    uids = []
    repo_uid = request.form.get("repo_uid")
    username = request.form.get("username")
    for key in request.form.keys():
        for value in request.form.getlist(key):
            if key.startswith("RSE-taxonomy"):
                uids.append(key)

    # Do the update for each criteria
    repo = app.client.get(repo_uid)
    print(f"Updating {repo} with {uids}")
    app.client.save_taxonomy(repo, username, uids)
    return username



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from .github import GitHubParser
from .gitlab import GitLabParser
from .zenodo import ZenodoParser
import re


def matches(Parser, uri):
    """Given a unique resource identifier, determine if it matches a regular expression
       used for a parser (or not)
    """
    if not hasattr(Parser, "matchstring"):
        raise NotImplementedError(f"{Parser.name} is missing a matchstring attribute.")

    return not re.search(Parser.matchstring, uri) == None


def get_parser(uri, config=None):
    """get parser will return the correct parser depending on a uri
    """
    parser = None
    if matches(GitHubParser, uri):
        parser = GitHubParser(uri)
    if matches(GitLabParser, uri):
        parser = GitLabParser(uri)
    if matches(ZenodoParser, uri):
        parser = ZenodoParser(uri)

    if not parser:
        raise NotImplementedError(f"There is no matching parser for {uri}")
    parser.config = config
    return parser


def get_named_parser(name, uri=None, config=None):
    """get a named parser, meaning determining based on name and not uri
    """
    parser = None
    if re.search("github", name):
        parser = GitHubParser(uri)
    elif re.search("gitlab", name):
        parser = GitLabParser(uri)
    elif re.search("zenodo", name):
        parser = ZenodoParser(uri)

    if not parser:
        raise NotImplementedError(f"There is no matching parser for {name}")

    parser.config = config
    return parser



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from rse.main.scrapers import get_named_scraper
import sys


def main(args, extra):

    try:
        scraper = get_named_scraper(args.scraper_name[0])
    except:
        sys.exit(f"{args.scraper_name[0]} is not a known scraper.")

    # Either get latest entries, or based on search
    if args.query is not None:
        results = scraper.search(args.query)
    else:
        results = scraper.latest()
    print("Found %s results" % len(results))

    # If we have results and it's not a dry run, create the repos
    if results and not args.dry_run:
        scraper.create(database=args.database, config_file=args.config_file)



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

import configparser
import logging
import os
import shutil
import sys

here = os.path.dirname(os.path.abspath(__file__))

bot = logging.getLogger("rse.main.config")


class Config:
    def __init__(self, config_file="rse.ini", load=True, generate=False):
        """Controller for an rse.ini config file.
        """
        self.configfile = os.path.abspath(config_file)

        # If the config file doesn't exist, generate it
        if not os.path.exists(self.configfile):
            if generate:
                default_config = os.path.join(here, "config.ini")
                bot.info(
                    "Generating configuration file %s"
                    % os.path.basename(self.configfile)
                )
                shutil.copyfile(default_config, self.configfile)
            else:
                sys.exit(
                    "rse.ini not found, specify with --config_file or move to directory."
                )

        # Load the config file if wanted
        if load and os.path.exists(self.configfile):
            self.read()

    def get(self, section, key):
        """A wrapper to config.get to directly interact with the self.config
        """
        return self.config.get(section, key)

    def update(self, section, key, value, save=False):
        """update client secrets will update the data structure for a particular
           authentication. This should only be used for a (quasi permanent) token
           or similar. The secrets file, if found, is updated and saved by default.
        """
        if section not in self.config:
            self.config[section] = {}
        self.config[section][key] = value

        # If save is true (to persist settings)
        if save:
            self.save()

    def save(self):
        """save configuration back to it's original file
        """
        with open(self.configfile, "w") as configfile:
            self.config.write(configfile)

    def read(self, configfile=None):
        """read in configuration file. By default use self.configfile.
        """
        configfile = configfile or self.configfile
        config = configparser.ConfigParser()
        config.read(configfile)
        self.config = config



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""


class Database:
    """A rse database holds common functions to represent repos, and metadata.
    """

    database = "notimplemented"

    def clear(self):
        """clear (delete) all software repositories.
        """
        raise NotImplementedError

    def get(self):
        """get a software repository from the database.
        """
        raise NotImplementedError

    def exists(self):
        """determine if a software repository exists in the database.
        """
        raise NotImplementedError

    def search(self, query):
        """search is only available to non-filesystem databases
        """
        raise NotImplementedError

    def update(self, uid):
        """update a software repository.
        """
        raise NotImplementedError

    def add(self, uid):
        """Add a new software repository to the database.
        """
        raise NotImplementedError

    def get_or_create(self, uid):
        """Determine if a repo exists.
        """
        raise NotImplementedError

    def delete_repo(self, uid):
        """delete a repo based on a specific identifier.
        """
        raise NotImplementedError

    def delete_parser(self, name):
        """delete all repos for a parser based on the parser name.
        """
        raise NotImplementedError

    def list_repos(self, name=None):
        """list software repositories, either under a particular parser name
           or just under all parsers.
        """
        raise NotImplementedError



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from rse.exceptions import DatabaseStringFormatError, MissingDatabaseString
from rse.main.config import Config
import logging
import os
import re


def main(args, extra):

    bot = logging.getLogger("rse.client")

    # The user wants to set the database
    if args.database:
        if not re.search("^(sqlite|filesystem|mysql+pymysql|postgres)", args.database):
            raise DatabaseStringFormatError

        # Ensure that database string, if exists, is correct
        database = args.database.split("://")[0]
        connection_string = args.database.replace(database, "", 1)

        # Instantiate a config with the correct configuration directory
        config = Config(config_file=args.config_file)

        # Update the database (not saved yet)
        config.update("DEFAULT", "database", database)

        # Mysql and postgres require connection strings
        if not connection_string and database in ["postgres", "mysql+pymysql"]:
            raise MissingDatabaseString(
                "You are required to provide a connection string for mysql+pymysql and postgres."
            )

        # Sqlite can be a path to a file, but must be located in root of config home
        elif connection_string and database == "sqlite":
            connection_string = os.path.basename(connection_string)

        # Connection string will be validated on init
        config.update("DEFAULT", "databaseConnect", connection_string)

        # If we make it here, save the configuration
        config.save()
        bot.info(f"Configuration saved with database {args.database}")

    # Set a value for a particular executor
    elif args.set:
        executor, key, value = args.set
        config = Config(config_dir=args.config_dir)
        executor = "executor.%s" % executor.lower()
        config.update(executor, key, value)
        config.save()
        bot.info(f"Configuration saved with {executor} {key} {value}")



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from flask import render_template
from rse.app.server import app
from rse.defaults import RSE_URL_PREFIX

## Main Index View


@app.route("/")
def index():

    return render_template(
        "home/filesystem-index.html",
        repos=app.client.list(),
        database=app.client.database,
        criteria=app.client.list_criteria(),
        entries=app.client.list_taxonomy(),
        url_prefix=RSE_URL_PREFIX,
        enable_annotate=not app.disable_annotate,
    )



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

import logging
import requests
import urllib.parse
from rse.utils.urls import check_response

from .base import ParserBase

bot = logging.getLogger("rse.main.parsers.gitlab")


class GitLabParser(ParserBase):

    name = "gitlab"
    matchstring = "gitlab"

    def __init__(self, uid=None, **kwargs):
        super().__init__(uid)

    def _set_uid(self, uid):
        """Given some kind of GitLab url, parse the uid
        """
        uid = uid.replace(":", "/")
        owner, repo = uid.replace(".git", "").split("/")[-2:]
        return "{}/{}".format(owner, repo)

    def load_secrets(self):
        """load secrets, namely the GitLab token
        """
        self.token = self.get_setting("TOKEN")

    def get_url(self, data=None):
        """a common function for a parser to return the html url for the
           upper level of metadata
        """
        data = data or self.data
        return data.get("web_url")

    def get_avatar(self, data=None):
        """a common function for a parser to return an image.
        """
        data = data or self.data
        return data.get("avatar_url") or "https://gitlab.com%s" % self.data.get(
            "namespace", {}
        ).get("avatar_url")

    def get_description(self, data=None):
        """a common function for a parser to return a description.
        """
        data = data or self.data
        return data.get("description")

    def get_metadata(self, uri=None):
        """Retrieve repository metadata. The common metadata (timestamp) is
           added by the software repository parser, and here we need to
           ensure that the url field is populated with a correct url.

           Arguments:
           uri (str) : a repository uri string to override one currently set
        """
        if uri:
            self.set_uri(uri)
        self.load_secrets()

        repo = "/".join(self.uid.split("/")[-2:])
        url = "https://gitlab.com/api/v4/projects/%s" % urllib.parse.quote(
            repo, safe=""
        )

        # Add authorization header if token is provided
        headers = None
        if self.token:
            headers = {"Authorization": "Bearer %s" % self.token}
        response = requests.get(url, headers=headers)

        # Successful query!
        self.data = check_response(response)
        return self.data



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

import os

## Important! This needs to not be dynamic if need to save sessions

WTF_CSRF_ENABLED = True
SECRET_KEY = os.environ.get("QME_SERVER_KEY", "banana-banana-fo-fanna")



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from rse.main import Encyclopedia
from rse.exceptions import DirectoryNotFoundError
import os


def main(args, extra):

    # present working directory
    path = args.path
    if args.path == ".":
        path = os.getcwd()

    # directory must exist!
    if not os.path.exists(path):
        raise DirectoryNotFoundError(path)
    config_file = os.path.join(path, args.config_file)

    # generate the software repository in the base
    Encyclopedia(config_file=config_file, generate=True, database=args.database)



#!/usr/bin/env python
"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

import os
import sys
import pytest


def test_parser_gitlab(tmp_path):
    """Test the gitlab parser.
    """
    from rse.main.parsers import GitLabParser

    for repo in [
        "singularityhub/gitlab-ci",
        "gitlab.com/singularityhub/gitlab-ci",
        "git@gitlab.com:singularityhub/gitlab-ci.git",
    ]:
        parser = GitLabParser(repo)
        assert parser.uid == "gitlab/singularityhub/gitlab-ci"
        assert parser.summary()

    # Only test one get of data
    assert parser.get_metadata()
    data = parser.export()
    for key in ["timestamp", "description", "web_url"]:
        assert key in data



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from rse.main import Encyclopedia
import logging

bot = logging.getLogger("rse.client")


def main(args, extra):

    enc = Encyclopedia(config_file=args.config_file)

    # Case 1: empty list indicates listing all
    if not args.uid:
        bot.error("Please provide a unique resource identifier to search for.")
    else:
        if enc.exists(args.uid):
            print(f"{args.uid} exists.")
        else:
            print(f"{args.uid} does not exist.")



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from rse.main import Encyclopedia
from rse.logger import bot
import sys


def main(args, extra):

    enc = Encyclopedia(config_file=args.config_file, database=args.database)
    query = " ".join(args.query).strip()
    if not query:
        sys.exit("Please provide a query to search for.")
    results = enc.search(query)
    if results:
        bot.table(results)



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

import random
import string


def main(args, extra):

    punctuation = "!#$%&()*+,-./:;<=>?@^_{|}~"
    choices = string.ascii_letters + string.digits + punctuation
    selected = [random.SystemRandom().choice(choices) for _ in range(50)]
    generated_key = "".join(selected)
    print(generated_key)
